@@ -2,8 +2,6 @@
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
-//go:build go1.20
-
 package main
 
 import (
@@ -17,23 +15,24 @@ import (
 	"go/types"
 	"io"
 	"log"
+	"maps"
 	"os"
 	"path/filepath"
 	"regexp"
 	"runtime"
 	"runtime/pprof"
+	"slices"
 	"sort"
 	"strings"
 	"text/template"
 
-	"golang.org/x/telemetry/counter"
-	"golang.org/x/telemetry/crashmonitor"
+	"golang.org/x/telemetry"
 	"golang.org/x/tools/go/callgraph"
 	"golang.org/x/tools/go/callgraph/rta"
 	"golang.org/x/tools/go/packages"
 	"golang.org/x/tools/go/ssa"
 	"golang.org/x/tools/go/ssa/ssautil"
-	"golang.org/x/tools/internal/aliases"
+	"golang.org/x/tools/internal/typesinternal"
 )
 
 //go:embed doc.go
@@ -65,8 +64,7 @@ Flags:
 }
 
 func main() {
-	counter.Open()       // Enable telemetry counter writing.
-	crashmonitor.Start() // Enable crash reporting watchdog.
+	telemetry.Start(telemetry.Config{ReportCrashes: true})
 
 	log.SetPrefix("deadcode: ")
 	log.SetFlags(0) // no time prefix
@@ -132,20 +130,19 @@ func main() {
 		log.Fatalf("packages contain errors")
 	}
 
-	// Gather names of generated files.
-	generated := make(map[string]bool)
-	packages.Visit(initial, nil, func(p *packages.Package) {
-		for _, file := range p.Syntax {
-			if isGenerated(file) {
-				generated[p.Fset.File(file.Pos()).Name()] = true
-			}
-		}
-	})
-
 	// If -filter is unset, use first module (if available).
 	if *filterFlag == "<module>" {
-		if mod := initial[0].Module; mod != nil && mod.Path != "" {
-			*filterFlag = "^" + regexp.QuoteMeta(mod.Path) + "\\b"
+		seen := make(map[string]bool)
+		var patterns []string
+		for _, pkg := range initial {
+			if pkg.Module != nil && pkg.Module.Path != "" && !seen[pkg.Module.Path] {
+				seen[pkg.Module.Path] = true
+				patterns = append(patterns, regexp.QuoteMeta(pkg.Module.Path))
+			}
+		}
+
+		if patterns != nil {
+			*filterFlag = "^(" + strings.Join(patterns, "|") + ")\\b"
 		} else {
 			*filterFlag = "" // match any
 		}
@@ -169,6 +166,46 @@ func main() {
 		roots = append(roots, main.Func("init"), main.Func("main"))
 	}
 
+	// Gather all source-level functions,
+	// as the user interface is expressed in terms of them.
+	//
+	// We ignore synthetic wrappers, and nested functions. Literal
+	// functions passed as arguments to other functions are of
+	// course address-taken and there exists a dynamic call of
+	// that signature, so when they are unreachable, it is
+	// invariably because the parent is unreachable.
+	var (
+		sourceFuncs    []*ssa.Function
+		generated      = make(map[string]bool)
+		interfaceTypes = make(map[*types.Package][]*types.Interface)
+	)
+	packages.Visit(initial, nil, func(p *packages.Package) {
+		// Collect interfaces by package for marker method identification.
+		var interfaces []*types.Interface
+		scope := p.Types.Scope()
+		for _, name := range scope.Names() {
+			if typeName, ok := scope.Lookup(name).(*types.TypeName); ok &&
+				types.IsInterface(typeName.Type()) {
+				interfaces = append(interfaces, typeName.Type().Underlying().(*types.Interface))
+			}
+		}
+		interfaceTypes[p.Types] = interfaces
+
+		for _, file := range p.Syntax {
+			for _, decl := range file.Decls {
+				if decl, ok := decl.(*ast.FuncDecl); ok {
+					obj := p.TypesInfo.Defs[decl.Name].(*types.Func)
+					fn := prog.FuncValue(obj)
+					sourceFuncs = append(sourceFuncs, fn)
+				}
+			}
+
+			if ast.IsGenerated(file) {
+				generated[p.Fset.File(file.Pos()).Name()] = true
+			}
+		}
+	})
+
 	// Compute the reachabilty from main.
 	// (Build a call graph only for -whylive.)
 	res := rta.Analyze(roots, *whyLiveFlag != "")
@@ -195,10 +232,7 @@ func main() {
 	// is not dead, by showing a path to it from some root.
 	if *whyLiveFlag != "" {
 		targets := make(map[*ssa.Function]bool)
-		for fn := range ssautil.AllFunctions(prog) {
-			if fn.Synthetic != "" || fn.Object() == nil {
-				continue // not a source-level named function
-			}
+		for _, fn := range sourceFuncs {
 			if prettyName(fn, true) == *whyLiveFlag {
 				targets[fn] = true
 			}
@@ -249,7 +283,7 @@ func main() {
 			edges = append(edges, jsonEdge{
 				Initial:  cond(len(edges) == 0, prettyName(edge.Caller.Func, true), ""),
 				Kind:     cond(isStaticCall(edge), "static", "dynamic"),
-				Position: toJSONPosition(prog.Fset.Position(edge.Site.Pos())),
+				Position: toJSONPosition(prog.Fset.Position(edge.Pos())),
 				Callee:   prettyName(edge.Callee.Func, true),
 			})
 		}
@@ -263,26 +297,7 @@ func main() {
 
 	// Group unreachable functions by package path.
 	byPkgPath := make(map[string]map[*ssa.Function]bool)
-	for fn := range ssautil.AllFunctions(prog) {
-		if fn.Synthetic != "" {
-			continue // ignore synthetic wrappers etc
-		}
-
-		// Use generic, as instantiations may not have a Pkg.
-		if orig := fn.Origin(); orig != nil {
-			fn = orig
-		}
-
-		// Ignore unreachable nested functions.
-		// Literal functions passed as arguments to other
-		// functions are of course address-taken and there
-		// exists a dynamic call of that signature, so when
-		// they are unreachable, it is invariably because the
-		// parent is unreachable.
-		if fn.Parent() != nil {
-			continue
-		}
-
+	for _, fn := range sourceFuncs {
 		posn := prog.Fset.Position(fn.Pos())
 
 		if !reachablePosn[posn] {
@@ -300,9 +315,7 @@ func main() {
 
 	// Build array of jsonPackage objects.
 	var packages []any
-	pkgpaths := keys(byPkgPath)
-	sort.Strings(pkgpaths)
-	for _, pkgpath := range pkgpaths {
+	for _, pkgpath := range slices.Sorted(maps.Keys(byPkgPath)) {
 		if !filter.MatchString(pkgpath) {
 			continue
 		}
@@ -313,7 +326,7 @@ func main() {
 		// declaration order. This tends to keep related
 		// methods such as (T).Marshal and (*T).Unmarshal
 		// together better than sorting.
-		fns := keys(m)
+		fns := slices.Collect(maps.Keys(m))
 		sort.Slice(fns, func(i, j int) bool {
 			xposn := prog.Fset.Position(fns[i].Pos())
 			yposn := prog.Fset.Position(fns[j].Pos())
@@ -335,10 +348,17 @@ func main() {
 				continue
 			}
 
+			// Marker methods should not be reported
+			marker := isMarkerMethod(fn, interfaceTypes[fn.Pkg.Pkg])
+			if marker {
+				continue
+			}
+
 			functions = append(functions, jsonFunction{
 				Name:      prettyName(fn, false),
 				Position:  toJSONPosition(posn),
 				Generated: gen,
+				Marker:    marker,
 			})
 		}
 		if len(functions) > 0 {
@@ -378,18 +398,15 @@ func prettyName(fn *ssa.Function, qualified bool) string {
 		// anonymous?
 		if fn.Parent() != nil {
 			format(fn.Parent())
-			i := index(fn.Parent().AnonFuncs, fn)
+			i := slices.Index(fn.Parent().AnonFuncs, fn)
 			fmt.Fprintf(&buf, "$%d", i+1)
 			return
 		}
 
 		// method receiver?
 		if recv := fn.Signature.Recv(); recv != nil {
-			t := recv.Type()
-			if ptr, ok := aliases.Unalias(t).(*types.Pointer); ok {
-				t = ptr.Elem()
-			}
-			buf.WriteString(aliases.Unalias(t).(*types.Named).Obj().Name())
+			_, named := typesinternal.ReceiverNamed(recv)
+			buf.WriteString(named.Obj().Name())
 			buf.WriteByte('.')
 		}
 
@@ -427,45 +444,6 @@ func printObjects(format string, objects []any) {
 	}
 }
 
-// TODO(adonovan): use go1.21's ast.IsGenerated.
-
-// isGenerated reports whether the file was generated by a program,
-// not handwritten, by detecting the special comment described
-// at https://go.dev/s/generatedcode.
-//
-// The syntax tree must have been parsed with the ParseComments flag.
-// Example:
-//
-//	f, err := parser.ParseFile(fset, filename, src, parser.ParseComments|parser.PackageClauseOnly)
-//	if err != nil { ... }
-//	gen := ast.IsGenerated(f)
-func isGenerated(file *ast.File) bool {
-	_, ok := generator(file)
-	return ok
-}
-
-func generator(file *ast.File) (string, bool) {
-	for _, group := range file.Comments {
-		for _, comment := range group.List {
-			if comment.Pos() > file.Package {
-				break // after package declaration
-			}
-			// opt: check Contains first to avoid unnecessary array allocation in Split.
-			const prefix = "// Code generated "
-			if strings.Contains(comment.Text, prefix) {
-				for _, line := range strings.Split(comment.Text, "\n") {
-					if rest, ok := strings.CutPrefix(line, prefix); ok {
-						if gen, ok := strings.CutSuffix(rest, " DO NOT EDIT."); ok {
-							return gen, true
-						}
-					}
-				}
-			}
-		}
-	}
-	return "", false
-}
-
 // pathSearch returns the shortest path from one of the roots to one
 // of the targets (along with the root itself), or zero if no path was found.
 func pathSearch(roots []*ssa.Function, res *rta.Result, targets map[*ssa.Function]bool) (*callgraph.Node, []*callgraph.Edge) {
@@ -479,7 +457,7 @@ func pathSearch(roots []*ssa.Function, res *rta.Result, targets map[*ssa.Functio
 	// Sort roots into preferred order.
 	importsTesting := func(fn *ssa.Function) bool {
 		isTesting := func(p *types.Package) bool { return p.Path() == "testing" }
-		return containsFunc(fn.Pkg.Pkg.Imports(), isTesting)
+		return slices.ContainsFunc(fn.Pkg.Pkg.Imports(), isTesting)
 	}
 	sort.Slice(roots, func(i, j int) bool {
 		x, y := roots[i], roots[j]
@@ -513,7 +491,7 @@ func pathSearch(roots []*ssa.Function, res *rta.Result, targets map[*ssa.Functio
 					for {
 						edge := seen[node]
 						if edge == nil {
-							reverse(path)
+							slices.Reverse(path)
 							return path
 						}
 						path = append(path, edge)
@@ -581,6 +559,30 @@ func cond[T any](cond bool, t, f T) T {
 	}
 }
 
+// isMarkerMethod reports whether fn is a marker method:
+// an unexported, empty-bodied method with no parameters or results
+// that implements some named interface type in the same package.
+func isMarkerMethod(fn *ssa.Function, interfaceTypes []*types.Interface) bool {
+	// Is it an unexported method of no params/results?
+	if !(fn.Signature.Recv() != nil &&
+		!ast.IsExported(fn.Name()) &&
+		fn.Signature.Params() == nil &&
+		fn.Signature.Results() == nil) {
+		return false
+	}
+
+	// Does the method have an empty body?
+	body := fn.Syntax().(*ast.FuncDecl).Body
+	if body == nil || len(body.List) > 0 {
+		return false
+	}
+
+	// Does it implement some named interface type in this package?
+	return slices.ContainsFunc(interfaceTypes, func(iface *types.Interface) bool {
+		return types.Implements(fn.Signature.Recv().Type(), iface)
+	})
+}
+
 // -- output protocol (for JSON or text/template) --
 
 // Keep in sync with doc comment!
@@ -589,6 +591,7 @@ type jsonFunction struct {
 	Name      string       // name (sans package qualifier)
 	Position  jsonPosition // file/line/column of declaration
 	Generated bool         // function is declared in a generated .go file
+	Marker    bool         // function is a marker interface method
 }
 
 func (f jsonFunction) String() string { return f.Name }
@@ -617,43 +620,3 @@ type jsonPosition struct {
 func (p jsonPosition) String() string {
 	return fmt.Sprintf("%s:%d:%d", p.File, p.Line, p.Col)
 }
-
-// -- from the future --
-
-// TODO(adonovan): use go1.22's slices and maps packages.
-
-func containsFunc[S ~[]E, E any](s S, f func(E) bool) bool {
-	return indexFunc(s, f) >= 0
-}
-
-func indexFunc[S ~[]E, E any](s S, f func(E) bool) int {
-	for i := range s {
-		if f(s[i]) {
-			return i
-		}
-	}
-	return -1
-}
-
-func index[S ~[]E, E comparable](s S, v E) int {
-	for i := range s {
-		if v == s[i] {
-			return i
-		}
-	}
-	return -1
-}
-
-func reverse[S ~[]E, E any](s S) {
-	for i, j := 0, len(s)-1; i < j; i, j = i+1, j-1 {
-		s[i], s[j] = s[j], s[i]
-	}
-}
-
-func keys[M ~map[K]V, K comparable, V any](m M) []K {
-	r := make([]K, 0, len(m))
-	for k := range m {
-		r = append(r, k)
-	}
-	return r
-}
