@@ -12,7 +12,6 @@ import (
 	"flag"
 	"fmt"
 	"log"
-	"math/rand"
 	"os"
 	"path/filepath"
 	"reflect"
@@ -28,10 +27,12 @@ import (
 	"golang.org/x/tools/gopls/internal/lsprpc"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/protocol/command"
+	"golang.org/x/tools/gopls/internal/protocol/semtok"
 	"golang.org/x/tools/gopls/internal/server"
 	"golang.org/x/tools/gopls/internal/settings"
 	"golang.org/x/tools/gopls/internal/util/browser"
 	bugpkg "golang.org/x/tools/gopls/internal/util/bug"
+	"golang.org/x/tools/gopls/internal/util/moreslices"
 	"golang.org/x/tools/internal/diff"
 	"golang.org/x/tools/internal/jsonrpc2"
 	"golang.org/x/tools/internal/tool"
@@ -62,8 +63,8 @@ type Application struct {
 	// VeryVerbose enables a higher level of verbosity in logging output.
 	VeryVerbose bool `flag:"vv,veryverbose" help:"very verbose output"`
 
-	// Control ocagent export of telemetry
-	OCAgent string `flag:"ocagent" help:"the address of the ocagent (e.g. http://localhost:55678), or off"`
+	// OTel specifies the OpenTelemetry collector endpoint (e.g. http://localhost:4318).
+	OTel string `flag:"otel" help:"export telemetry to specified OpenTelemetry collector address (e.g. http://localhost:4318)"`
 
 	// PrepareOptions is called to update the options when a new view is built.
 	// It is primarily to allow the behavior of gopls to be modified by hooks.
@@ -97,8 +98,6 @@ func (app *Application) verbose() bool {
 // New returns a new Application ready to run.
 func New() *Application {
 	app := &Application{
-		OCAgent: "off", //TODO: Remove this line to default the exporter to on
-
 		Serve: Serve{
 			RemoteListenTimeout: 1 * time.Minute,
 		},
@@ -219,7 +218,7 @@ func isZeroValue(f *flag.Flag, value string) bool {
 	// This works unless the Value type is itself an interface type.
 	typ := reflect.TypeOf(f.Value)
 	var z reflect.Value
-	if typ.Kind() == reflect.Ptr {
+	if typ.Kind() == reflect.Pointer {
 		z = reflect.New(typ.Elem())
 	} else {
 		z = reflect.Zero(typ)
@@ -237,7 +236,7 @@ func (app *Application) Run(ctx context.Context, args ...string) error {
 	// executable, and immediately runs a gc.
 	filecache.Start()
 
-	ctx = debug.WithInstance(ctx, app.OCAgent)
+	ctx = debug.WithInstance(ctx, app.OTel)
 	if len(args) == 0 {
 		s := flag.NewFlagSet(app.Name(), flag.ExitOnError)
 		return tool.Run(ctx, s, &app.Serve, args)
@@ -283,7 +282,7 @@ func (app *Application) internalCommands() []tool.Application {
 func (app *Application) featureCommands() []tool.Application {
 	return []tool.Application{
 		&callHierarchy{app: app},
-		&check{app: app},
+		&check{app: app, Severity: "warning"},
 		&codeaction{app: app},
 		&codelens{app: app},
 		&definition{app: app},
@@ -291,6 +290,7 @@ func (app *Application) featureCommands() []tool.Application {
 		&fix{app: app}, // (non-functional)
 		&foldingRanges{app: app},
 		&format{app: app},
+		&headlessMCP{app: app},
 		&highlight{app: app},
 		&implementation{app: app},
 		&imports{app: app},
@@ -300,7 +300,7 @@ func (app *Application) featureCommands() []tool.Application {
 		&prepareRename{app: app},
 		&references{app: app},
 		&rename{app: app},
-		&semtok{app: app},
+		&semanticToken{app: app},
 		&signature{app: app},
 		&stats{app: app},
 		&symbols{app: app},
@@ -309,26 +309,28 @@ func (app *Application) featureCommands() []tool.Application {
 	}
 }
 
-var (
-	internalMu          sync.Mutex
-	internalConnections = make(map[string]*connection)
-)
-
-// connect creates and initializes a new in-process gopls session.
-func (app *Application) connect(ctx context.Context) (*connection, error) {
+// connect creates and initializes a new in-process gopls LSP session.
+func (app *Application) connect(ctx context.Context) (*client, *cache.Session, error) {
+	root, err := os.Getwd()
+	if err != nil {
+		return nil, nil, fmt.Errorf("finding workdir: %v", err)
+	}
+	options := settings.DefaultOptions(app.options)
 	client := newClient(app)
-	var svr protocol.Server
+	var (
+		svr  protocol.Server
+		sess *cache.Session
+	)
 	if app.Remote == "" {
 		// local
-		options := settings.DefaultOptions(app.options)
-		svr = server.New(cache.NewSession(ctx, cache.New(nil)), client, options)
+		sess = cache.NewSession(ctx, cache.New(nil))
+		svr = server.New(sess, client, options)
 		ctx = protocol.WithClient(ctx, client)
-
 	} else {
 		// remote
 		netConn, err := lsprpc.ConnectToRemote(ctx, app.Remote)
 		if err != nil {
-			return nil, err
+			return nil, nil, err
 		}
 		stream := jsonrpc2.NewHeaderStream(netConn)
 		jsonConn := jsonrpc2.NewConn(stream)
@@ -338,22 +340,19 @@ func (app *Application) connect(ctx context.Context) (*connection, error) {
 			protocol.Handlers(
 				protocol.ClientHandler(client, jsonrpc2.MethodNotFound)))
 	}
-	conn := newConnection(svr, client)
-	return conn, conn.initialize(ctx, app.options)
+	if err := client.initialize(ctx, svr, initParams(root, options)); err != nil {
+		return nil, nil, err
+	}
+	return client, sess, nil
 }
 
-func (c *connection) initialize(ctx context.Context, options func(*settings.Options)) error {
-	wd, err := os.Getwd()
-	if err != nil {
-		return fmt.Errorf("finding workdir: %v", err)
-	}
+func initParams(rootDir string, opts *settings.Options) *protocol.ParamInitialize {
 	params := &protocol.ParamInitialize{}
-	params.RootURI = protocol.URIFromPath(wd)
+	params.RootURI = protocol.URIFromPath(rootDir)
 	params.Capabilities.Workspace.Configuration = true
 
-	// Make sure to respect configured options when sending initialize request.
-	opts := settings.DefaultOptions(options)
-	// If you add an additional option here, you must update the map key in connect.
+	// If you add an additional option here,
+	// you must update the map key of settings.DefaultOptions called in (*Application).connect.
 	params.Capabilities.TextDocument.Hover = &protocol.HoverClientCapabilities{
 		ContentFormat: []protocol.MarkupKind{opts.PreferredContentFormat},
 	}
@@ -361,63 +360,62 @@ func (c *connection) initialize(ctx context.Context, options func(*settings.Opti
 	params.Capabilities.TextDocument.SemanticTokens = protocol.SemanticTokensClientCapabilities{}
 	params.Capabilities.TextDocument.SemanticTokens.Formats = []protocol.TokenFormat{"relative"}
 	params.Capabilities.TextDocument.SemanticTokens.Requests.Range = &protocol.Or_ClientSemanticTokensRequestOptions_range{Value: true}
-	//params.Capabilities.TextDocument.SemanticTokens.Requests.Range.Value = true
+	// params.Capabilities.TextDocument.SemanticTokens.Requests.Range.Value = true
 	params.Capabilities.TextDocument.SemanticTokens.Requests.Full = &protocol.Or_ClientSemanticTokensRequestOptions_full{Value: true}
-	params.Capabilities.TextDocument.SemanticTokens.TokenTypes = protocol.SemanticTypes()
-	params.Capabilities.TextDocument.SemanticTokens.TokenModifiers = protocol.SemanticModifiers()
+	params.Capabilities.TextDocument.SemanticTokens.TokenTypes = moreslices.ConvertStrings[string](semtok.Types)
+	params.Capabilities.TextDocument.SemanticTokens.TokenModifiers = moreslices.ConvertStrings[string](semtok.Modifiers)
+	params.Capabilities.TextDocument.CodeAction = protocol.CodeActionClientCapabilities{
+		CodeActionLiteralSupport: protocol.ClientCodeActionLiteralOptions{
+			CodeActionKind: protocol.ClientCodeActionKindOptions{
+				ValueSet: []protocol.CodeActionKind{protocol.Empty}, // => all
+			},
+		},
+	}
 	params.Capabilities.Window.WorkDoneProgress = true
-
-	params.InitializationOptions = map[string]interface{}{
+	params.Capabilities.Workspace.FileOperations = &protocol.FileOperationClientCapabilities{
+		DidCreate: true,
+	}
+	params.InitializationOptions = map[string]any{
 		"symbolMatcher": string(opts.SymbolMatcher),
 	}
-	if _, err := c.Server.Initialize(ctx, params); err != nil {
+	return params
+}
+
+// initialize performs LSP's two-call client/server handshake.
+func (cli *client) initialize(ctx context.Context, server protocol.Server, params *protocol.ParamInitialize) error {
+	result, err := server.Initialize(ctx, params)
+	if err != nil {
 		return err
 	}
-	if err := c.Server.Initialized(ctx, &protocol.InitializedParams{}); err != nil {
+	if err := server.Initialized(ctx, &protocol.InitializedParams{}); err != nil {
 		return err
 	}
+	cli.server = server
+	cli.initializeResult = result
 	return nil
 }
 
-type connection struct {
-	protocol.Server
-	client *cmdClient
-}
-
-// registerProgressHandler registers a handler for progress notifications.
-// The caller must call unregister when the handler is no longer needed.
-func (cli *cmdClient) registerProgressHandler(handler func(*protocol.ProgressParams)) (token protocol.ProgressToken, unregister func()) {
-	token = fmt.Sprintf("tok%d", rand.Uint64())
-
-	// register
-	cli.progressHandlersMu.Lock()
-	if cli.progressHandlers == nil {
-		cli.progressHandlers = make(map[protocol.ProgressToken]func(*protocol.ProgressParams))
-	}
-	cli.progressHandlers[token] = handler
-	cli.progressHandlersMu.Unlock()
-
-	unregister = func() {
-		cli.progressHandlersMu.Lock()
-		delete(cli.progressHandlers, token)
-		cli.progressHandlersMu.Unlock()
-	}
-	return token, unregister
-}
-
-// cmdClient defines the protocol.Client interface behavior of the gopls CLI tool.
-type cmdClient struct {
+// client implements [protocol.Client] and defines the LSP client
+// operations of the gopls command.
+//
+// It holds the client-side state of a single client/server
+// connection; it conceptually corresponds to a single call to
+// connect(2).
+type client struct {
 	app *Application
 
-	progressHandlersMu sync.Mutex
-	progressHandlers   map[protocol.ProgressToken]func(*protocol.ProgressParams)
-	iwlToken           protocol.ProgressToken
-	iwlDone            chan struct{}
+	server           protocol.Server
+	initializeResult *protocol.InitializeResult // includes server capabilities
+
+	progressMu sync.Mutex
+	iwlToken   protocol.ProgressToken
+	iwlDone    chan struct{}
 
 	filesMu sync.Mutex // guards files map
 	files   map[protocol.DocumentURI]*cmdFile
 }
 
+// cmdFile represents an open file in the gopls command LSP client.
 type cmdFile struct {
 	uri           protocol.DocumentURI
 	mapper        *protocol.Mapper
@@ -426,37 +424,34 @@ type cmdFile struct {
 	diagnostics   []protocol.Diagnostic
 }
 
-func newClient(app *Application) *cmdClient {
-	return &cmdClient{
+func newClient(app *Application) *client {
+	return &client{
 		app:     app,
 		files:   make(map[protocol.DocumentURI]*cmdFile),
 		iwlDone: make(chan struct{}),
 	}
 }
 
-func newConnection(server protocol.Server, client *cmdClient) *connection {
-	return &connection{
-		Server: server,
-		client: client,
-	}
+func (cli *client) TextDocumentContentRefresh(context.Context, *protocol.TextDocumentContentRefreshParams) error {
+	return nil
 }
 
-func (c *cmdClient) CodeLensRefresh(context.Context) error { return nil }
+func (cli *client) CodeLensRefresh(context.Context) error { return nil }
 
-func (c *cmdClient) FoldingRangeRefresh(context.Context) error { return nil }
+func (cli *client) FoldingRangeRefresh(context.Context) error { return nil }
 
-func (c *cmdClient) LogTrace(context.Context, *protocol.LogTraceParams) error { return nil }
+func (cli *client) LogTrace(context.Context, *protocol.LogTraceParams) error { return nil }
 
-func (c *cmdClient) ShowMessage(ctx context.Context, p *protocol.ShowMessageParams) error {
+func (cli *client) ShowMessage(ctx context.Context, p *protocol.ShowMessageParams) error {
 	fmt.Fprintf(os.Stderr, "%s: %s\n", p.Type, p.Message)
 	return nil
 }
 
-func (c *cmdClient) ShowMessageRequest(ctx context.Context, p *protocol.ShowMessageRequestParams) (*protocol.MessageActionItem, error) {
+func (cli *client) ShowMessageRequest(ctx context.Context, p *protocol.ShowMessageRequestParams) (*protocol.MessageActionItem, error) {
 	return nil, nil
 }
 
-func (c *cmdClient) LogMessage(ctx context.Context, p *protocol.LogMessageParams) error {
+func (cli *client) LogMessage(ctx context.Context, p *protocol.LogMessageParams) error {
 	// This logic causes server logging to be double-prefixed with a timestamp.
 	//     2023/11/08 10:50:21 Error:2023/11/08 10:50:21 <actual message>
 	// TODO(adonovan): print just p.Message, plus a newline if needed?
@@ -466,42 +461,42 @@ func (c *cmdClient) LogMessage(ctx context.Context, p *protocol.LogMessageParams
 	case protocol.Warning:
 		log.Print("Warning:", p.Message)
 	case protocol.Info:
-		if c.app.verbose() {
+		if cli.app.verbose() {
 			log.Print("Info:", p.Message)
 		}
 	case protocol.Log:
-		if c.app.verbose() {
+		if cli.app.verbose() {
 			log.Print("Log:", p.Message)
 		}
 	default:
-		if c.app.verbose() {
+		if cli.app.verbose() {
 			log.Print(p.Message)
 		}
 	}
 	return nil
 }
 
-func (c *cmdClient) Event(ctx context.Context, t *interface{}) error { return nil }
+func (cli *client) Event(ctx context.Context, t *any) error { return nil }
 
-func (c *cmdClient) RegisterCapability(ctx context.Context, p *protocol.RegistrationParams) error {
+func (cli *client) RegisterCapability(ctx context.Context, p *protocol.RegistrationParams) error {
 	return nil
 }
 
-func (c *cmdClient) UnregisterCapability(ctx context.Context, p *protocol.UnregistrationParams) error {
+func (cli *client) UnregisterCapability(ctx context.Context, p *protocol.UnregistrationParams) error {
 	return nil
 }
 
-func (c *cmdClient) WorkspaceFolders(ctx context.Context) ([]protocol.WorkspaceFolder, error) {
+func (cli *client) WorkspaceFolders(ctx context.Context) ([]protocol.WorkspaceFolder, error) {
 	return nil, nil
 }
 
-func (c *cmdClient) Configuration(ctx context.Context, p *protocol.ParamConfiguration) ([]interface{}, error) {
-	results := make([]interface{}, len(p.Items))
+func (cli *client) Configuration(ctx context.Context, p *protocol.ParamConfiguration) ([]any, error) {
+	results := make([]any, len(p.Items))
 	for i, item := range p.Items {
 		if item.Section != "gopls" {
 			continue
 		}
-		m := map[string]interface{}{
+		m := map[string]any{
 			"analyses": map[string]any{
 				"fillreturns":    true,
 				"nonewvars":      true,
@@ -509,7 +504,7 @@ func (c *cmdClient) Configuration(ctx context.Context, p *protocol.ParamConfigur
 				"undeclaredname": true,
 			},
 		}
-		if c.app.VeryVerbose {
+		if cli.app.VeryVerbose {
 			m["verboseOutput"] = true
 		}
 		results[i] = m
@@ -517,8 +512,8 @@ func (c *cmdClient) Configuration(ctx context.Context, p *protocol.ParamConfigur
 	return results, nil
 }
 
-func (c *cmdClient) ApplyEdit(ctx context.Context, p *protocol.ApplyWorkspaceEditParams) (*protocol.ApplyWorkspaceEditResult, error) {
-	if err := c.applyWorkspaceEdit(&p.Edit); err != nil {
+func (cli *client) ApplyEdit(ctx context.Context, p *protocol.ApplyWorkspaceEditParams) (*protocol.ApplyWorkspaceEditResult, error) {
+	if err := cli.applyWorkspaceEdit(&p.Edit); err != nil {
 		return &protocol.ApplyWorkspaceEditResult{FailureReason: err.Error()}, nil
 	}
 	return &protocol.ApplyWorkspaceEditResult{Applied: true}, nil
@@ -531,7 +526,7 @@ func (c *cmdClient) ApplyEdit(ctx context.Context, p *protocol.ApplyWorkspaceEdi
 // See also:
 //   - changedFiles in ../test/marker/marker_test.go for the golden-file capturing variant
 //   - applyWorkspaceEdit in ../test/integration/fake/editor.go for the Editor variant
-func (cli *cmdClient) applyWorkspaceEdit(wsedit *protocol.WorkspaceEdit) error {
+func (cli *client) applyWorkspaceEdit(wsedit *protocol.WorkspaceEdit) error {
 
 	create := func(uri protocol.DocumentURI, content []byte) error {
 		edits := []diff.Edit{{Start: 0, End: 0, New: string(content)}}
@@ -546,7 +541,7 @@ func (cli *cmdClient) applyWorkspaceEdit(wsedit *protocol.WorkspaceEdit) error {
 	for _, c := range wsedit.DocumentChanges {
 		switch {
 		case c.TextDocumentEdit != nil:
-			f := cli.openFile(c.TextDocumentEdit.TextDocument.URI)
+			f := cli.getFile(c.TextDocumentEdit.TextDocument.URI)
 			if f.err != nil {
 				return f.err
 			}
@@ -563,7 +558,7 @@ func (cli *cmdClient) applyWorkspaceEdit(wsedit *protocol.WorkspaceEdit) error {
 
 		case c.RenameFile != nil:
 			// Analyze as creation + deletion. (NB: loses file mode.)
-			f := cli.openFile(c.RenameFile.OldURI)
+			f := cli.getFile(c.RenameFile.OldURI)
 			if f.err != nil {
 				return f.err
 			}
@@ -575,7 +570,7 @@ func (cli *cmdClient) applyWorkspaceEdit(wsedit *protocol.WorkspaceEdit) error {
 			}
 
 		case c.DeleteFile != nil:
-			f := cli.openFile(c.DeleteFile.URI)
+			f := cli.getFile(c.DeleteFile.URI)
 			if f.err != nil {
 				return f.err
 			}
@@ -653,25 +648,28 @@ func updateFile(filename string, old, new []byte, edits []diff.Edit, flags *Edit
 	return nil
 }
 
-func (c *cmdClient) PublishDiagnostics(ctx context.Context, p *protocol.PublishDiagnosticsParams) error {
+func (cli *client) PublishDiagnostics(ctx context.Context, p *protocol.PublishDiagnosticsParams) error {
 	// Don't worry about diagnostics without versions.
+	//
+	// (Note: the representation of PublishDiagnosticsParams
+	// cannot distinguish a missing Version from v0, but the
+	// server never sends back an explicit zero.)
 	if p.Version == 0 {
 		return nil
 	}
 
-	c.filesMu.Lock()
-	file := c.getFile(p.URI)
-	c.filesMu.Unlock()
+	file := cli.getFile(p.URI)
 
 	file.diagnosticsMu.Lock()
 	defer file.diagnosticsMu.Unlock()
+
 	file.diagnostics = append(file.diagnostics, p.Diagnostics...)
 
 	// Perform a crude in-place deduplication.
 	// TODO(golang/go#60122): replace the gopls.diagnose_files
 	// command with support for textDocument/diagnostic,
 	// so that we don't need to do this de-duplication.
-	type key [6]interface{}
+	type key [6]any
 	seen := make(map[key]bool)
 	out := file.diagnostics[:0]
 	for _, d := range file.diagnostics {
@@ -690,52 +688,44 @@ func (c *cmdClient) PublishDiagnostics(ctx context.Context, p *protocol.PublishD
 	return nil
 }
 
-func (c *cmdClient) Progress(_ context.Context, params *protocol.ProgressParams) error {
-	token, ok := params.Token.(string)
-	if !ok {
+func (cli *client) Progress(_ context.Context, params *protocol.ProgressParams) error {
+	if _, ok := params.Token.(string); !ok {
 		return fmt.Errorf("unexpected progress token: %[1]T %[1]v", params.Token)
 	}
 
-	c.progressHandlersMu.Lock()
-	handler := c.progressHandlers[token]
-	c.progressHandlersMu.Unlock()
-	if handler == nil {
-		handler = c.defaultProgressHandler
-	}
-	handler(params)
-	return nil
-}
-
-// defaultProgressHandler is the default handler of progress messages,
-// used during the initialize request.
-func (c *cmdClient) defaultProgressHandler(params *protocol.ProgressParams) {
 	switch v := params.Value.(type) {
 	case *protocol.WorkDoneProgressBegin:
 		if v.Title == server.DiagnosticWorkTitle(server.FromInitialWorkspaceLoad) {
-			c.progressHandlersMu.Lock()
-			c.iwlToken = params.Token
-			c.progressHandlersMu.Unlock()
+			cli.progressMu.Lock()
+			cli.iwlToken = params.Token
+			cli.progressMu.Unlock()
+		}
+
+	case *protocol.WorkDoneProgressReport:
+		if cli.app.Verbose {
+			fmt.Fprintln(os.Stderr, v.Message)
 		}
 
 	case *protocol.WorkDoneProgressEnd:
-		c.progressHandlersMu.Lock()
-		iwlToken := c.iwlToken
-		c.progressHandlersMu.Unlock()
+		cli.progressMu.Lock()
+		iwlToken := cli.iwlToken
+		cli.progressMu.Unlock()
 
 		if params.Token == iwlToken {
-			close(c.iwlDone)
+			close(cli.iwlDone)
 		}
 	}
+	return nil
 }
 
-func (c *cmdClient) ShowDocument(ctx context.Context, params *protocol.ShowDocumentParams) (*protocol.ShowDocumentResult, error) {
+func (cli *client) ShowDocument(ctx context.Context, params *protocol.ShowDocumentParams) (*protocol.ShowDocumentResult, error) {
 	var success bool
 	if params.External {
 		// Open URI in external browser.
 		success = browser.Open(params.URI)
 	} else {
 		// Open file in editor, optionally taking focus and selecting a range.
-		// (cmdClient has no editor. Should it fork+exec $EDITOR?)
+		// (client has no editor. Should it fork+exec $EDITOR?)
 		log.Printf("Server requested that client editor open %q (takeFocus=%t, selection=%+v)",
 			params.URI, params.TakeFocus, params.Selection)
 		success = true
@@ -743,33 +733,37 @@ func (c *cmdClient) ShowDocument(ctx context.Context, params *protocol.ShowDocum
 	return &protocol.ShowDocumentResult{Success: success}, nil
 }
 
-func (c *cmdClient) WorkDoneProgressCreate(context.Context, *protocol.WorkDoneProgressCreateParams) error {
+func (cli *client) WorkDoneProgressCreate(context.Context, *protocol.WorkDoneProgressCreateParams) error {
 	return nil
 }
 
-func (c *cmdClient) DiagnosticRefresh(context.Context) error {
+func (cli *client) DiagnosticRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) InlayHintRefresh(context.Context) error {
+func (cli *client) InlayHintRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) SemanticTokensRefresh(context.Context) error {
+func (cli *client) SemanticTokensRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) InlineValueRefresh(context.Context) error {
+func (cli *client) InlineValueRefresh(context.Context) error {
 	return nil
 }
 
-func (c *cmdClient) getFile(uri protocol.DocumentURI) *cmdFile {
-	file, found := c.files[uri]
+// getFile returns the specified file, adding it to the client state if needed.
+func (cli *client) getFile(uri protocol.DocumentURI) *cmdFile {
+	cli.filesMu.Lock()
+	defer cli.filesMu.Unlock()
+
+	file, found := cli.files[uri]
 	if !found || file.err != nil {
 		file = &cmdFile{
 			uri: uri,
 		}
-		c.files[uri] = file
+		cli.files[uri] = file
 	}
 	if file.mapper == nil {
 		content, err := os.ReadFile(uri.Path())
@@ -782,30 +776,38 @@ func (c *cmdClient) getFile(uri protocol.DocumentURI) *cmdFile {
 	return file
 }
 
-func (c *cmdClient) openFile(uri protocol.DocumentURI) *cmdFile {
-	c.filesMu.Lock()
-	defer c.filesMu.Unlock()
-	return c.getFile(uri)
-}
-
-// TODO(adonovan): provide convenience helpers to:
-// - map a (URI, protocol.Range) to a MappedRange;
-// - parse a command-line argument to a MappedRange.
-func (c *connection) openFile(ctx context.Context, uri protocol.DocumentURI) (*cmdFile, error) {
-	file := c.client.openFile(uri)
+// openFile returns the specified file, adding it to the client state
+// if needed, and notifying the server that it was opened.
+func (cli *client) openFile(ctx context.Context, uri protocol.DocumentURI) (*cmdFile, error) {
+	file := cli.getFile(uri)
 	if file.err != nil {
 		return nil, file.err
 	}
 
+	// Choose language ID from file extension.
+	var langID protocol.LanguageKind // "" eventually maps to file.UnknownKind
+	switch filepath.Ext(uri.Path()) {
+	case ".go":
+		langID = "go"
+	case ".mod":
+		langID = "go.mod"
+	case ".sum":
+		langID = "go.sum"
+	case ".work":
+		langID = "go.work"
+	case ".s":
+		langID = "go.s"
+	}
+
 	p := &protocol.DidOpenTextDocumentParams{
 		TextDocument: protocol.TextDocumentItem{
 			URI:        uri,
-			LanguageID: "go",
+			LanguageID: langID,
 			Version:    1,
 			Text:       string(file.mapper.Content),
 		},
 	}
-	if err := c.Server.DidOpen(ctx, p); err != nil {
+	if err := cli.server.DidOpen(ctx, p); err != nil {
 		// TODO(adonovan): is this assignment concurrency safe?
 		file.err = fmt.Errorf("%v: %v", uri, err)
 		return nil, file.err
@@ -813,35 +815,26 @@ func (c *connection) openFile(ctx context.Context, uri protocol.DocumentURI) (*c
 	return file, nil
 }
 
-func (c *connection) semanticTokens(ctx context.Context, p *protocol.SemanticTokensRangeParams) (*protocol.SemanticTokens, error) {
-	// use range to avoid limits on full
-	resp, err := c.Server.SemanticTokensRange(ctx, p)
-	if err != nil {
-		return nil, err
-	}
-	return resp, nil
-}
-
-func (c *connection) diagnoseFiles(ctx context.Context, files []protocol.DocumentURI) error {
-	cmd, err := command.NewDiagnoseFilesCommand("Diagnose files", command.DiagnoseFilesArgs{
+func diagnoseFiles(ctx context.Context, server protocol.Server, files []protocol.DocumentURI) error {
+	cmd := command.NewDiagnoseFilesCommand("Diagnose files", command.DiagnoseFilesArgs{
 		Files: files,
 	})
-	if err != nil {
-		return err
-	}
-	_, err = c.executeCommand(ctx, &cmd)
+	_, err := executeCommand(ctx, server, cmd)
 	return err
 }
 
-func (c *connection) terminate(ctx context.Context) {
-	//TODO: do we need to handle errors on these calls?
-	c.Shutdown(ctx)
-	//TODO: right now calling exit terminates the process, we should rethink that
-	//server.Exit(ctx)
+func (cli *client) terminate(ctx context.Context) {
+	if err := cli.server.Shutdown(ctx); err != nil {
+		log.Printf("server shutdown failed: %v", err)
+	}
+
+	// Don't call Exit as it terminates the server process,
+	// which is the same as this client process.
+	// c.server.Exit(ctx)
 }
 
 // Implement io.Closer.
-func (c *cmdClient) Close() error {
+func (cli *client) Close() error {
 	return nil
 }
 
@@ -900,7 +893,7 @@ func (f *cmdFile) spanLocation(s span) (protocol.Location, error) {
 	if err != nil {
 		return protocol.Location{}, err
 	}
-	return f.mapper.RangeLocation(rng), nil
+	return f.mapper.URI.Location(rng), nil
 }
 
 // spanRange converts a (UTF-8) span to a protocol (UTF-16) range.
@@ -912,7 +905,7 @@ func (f *cmdFile) spanRange(s span) (protocol.Range, error) {
 	// case-sensitive directories. The authoritative answer
 	// requires querying the file system, and we don't want
 	// to do that.
-	if !strings.EqualFold(filepath.Base(string(f.mapper.URI)), filepath.Base(string(s.URI()))) {
+	if !strings.EqualFold(f.mapper.URI.Base(), s.URI().Base()) {
 		return protocol.Range{}, bugpkg.Errorf("mapper is for file %q instead of %q", f.mapper.URI, s.URI())
 	}
 	start, err := pointPosition(f.mapper, s.Start())
