@@ -25,15 +25,18 @@ import (
 	"time"
 
 	"golang.org/x/tools/gopls/internal/cache"
+	"golang.org/x/tools/gopls/internal/cache/metadata"
 	"golang.org/x/tools/gopls/internal/debug/log"
+	"golang.org/x/tools/gopls/internal/file"
 	label1 "golang.org/x/tools/gopls/internal/label"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/util/bug"
+	versionpkg "golang.org/x/tools/gopls/internal/version"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/event/core"
 	"golang.org/x/tools/internal/event/export"
 	"golang.org/x/tools/internal/event/export/metric"
-	"golang.org/x/tools/internal/event/export/ocagent"
+	"golang.org/x/tools/internal/event/export/otel"
 	"golang.org/x/tools/internal/event/export/prometheus"
 	"golang.org/x/tools/internal/event/keys"
 	"golang.org/x/tools/internal/event/label"
@@ -51,16 +54,15 @@ type Instance struct {
 	Logfile       string
 	StartTime     time.Time
 	ServerAddress string
-	OCAgentConfig string
 
 	LogWriter io.Writer
 
 	exporter event.Exporter
 
-	ocagent    *ocagent.Exporter
 	prometheus *prometheus.Exporter
 	rpcs       *Rpcs
 	traces     *traces
+	otel       *otel.Exporter
 	State      *State
 
 	serveMu              sync.Mutex
@@ -141,12 +143,12 @@ func (st *State) Views() []*cache.View {
 
 // View returns the View that matches the supplied id.
 func (st *State) View(id string) *cache.View {
-	for _, v := range st.Views() {
-		if v.ID() == id {
+	for _, s := range st.Sessions() {
+		if v, err := s.View(id); err == nil {
 			return v
 		}
 	}
-	return nil
+	return nil // not found
 }
 
 // Clients returns the set of Clients currently being served.
@@ -280,23 +282,23 @@ func cmdline(w http.ResponseWriter, r *http.Request) {
 	pprof.Cmdline(fake, r)
 }
 
-func (i *Instance) getCache(r *http.Request) interface{} {
+func (i *Instance) getCache(r *http.Request) any {
 	return i.State.Cache(path.Base(r.URL.Path))
 }
 
-func (i *Instance) getAnalysis(r *http.Request) interface{} {
+func (i *Instance) getAnalysis(r *http.Request) any {
 	return i.State.Analysis()
 }
 
-func (i *Instance) getSession(r *http.Request) interface{} {
+func (i *Instance) getSession(r *http.Request) any {
 	return i.State.Session(path.Base(r.URL.Path))
 }
 
-func (i *Instance) getClient(r *http.Request) interface{} {
+func (i *Instance) getClient(r *http.Request) any {
 	return i.State.Client(path.Base(r.URL.Path))
 }
 
-func (i *Instance) getServer(r *http.Request) interface{} {
+func (i *Instance) getServer(r *http.Request) any {
 	i.State.mu.Lock()
 	defer i.State.mu.Unlock()
 	id := path.Base(r.URL.Path)
@@ -308,13 +310,18 @@ func (i *Instance) getServer(r *http.Request) interface{} {
 	return nil
 }
 
-func (i *Instance) getFile(r *http.Request) interface{} {
-	identifier := path.Base(r.URL.Path)
-	sid := path.Base(path.Dir(r.URL.Path))
-	s := i.State.Session(sid)
+type FileWithKind interface {
+	file.Handle
+	Kind() file.Kind // (overlay files only)
+}
+
+// /file/{session}/{identifier}. Returns a [FileWithKind].
+func (i *Instance) getFile(r *http.Request) any {
+	s := i.State.Session(r.PathValue("session"))
 	if s == nil {
 		return nil
 	}
+	identifier := r.PathValue("identifier")
 	for _, o := range s.Overlays() {
 		// TODO(adonovan): understand and document this comparison.
 		if o.Identity().Hash.String() == identifier {
@@ -324,9 +331,94 @@ func (i *Instance) getFile(r *http.Request) interface{} {
 	return nil
 }
 
-func (i *Instance) getInfo(r *http.Request) interface{} {
+type MetadataInfo struct {
+	SessionID string
+	ViewID    string
+	Graph     *metadata.Graph
+}
+
+// /metadata/{session}/{view}. Returns a [*MetadataInfo].
+func (i *Instance) getMetadata(r *http.Request) any {
+	session := i.State.Session(r.PathValue("session"))
+	if session == nil {
+		return nil
+	}
+
+	v, err := session.View(r.PathValue("view"))
+	if err != nil {
+		stdlog.Printf("/metadata: %v", err)
+		return nil // no found
+	}
+
+	snapshot, release, err := v.Snapshot()
+	if err != nil {
+		stdlog.Printf("/metadata: failed to get latest snapshot: %v", err)
+		return nil
+	}
+	defer release()
+	return &MetadataInfo{
+		SessionID: session.ID(),
+		ViewID:    v.ID(),
+		Graph:     snapshot.MetadataGraph(),
+	}
+}
+
+type PackageInfo struct {
+	SessionID   string
+	ViewID      string
+	Package     *cache.Package
+	Diagnostics map[protocol.DocumentURI][]*cache.Diagnostic
+}
+
+// /package/{session}/{view}/{id...}. Returns a [*PackageInfo].
+func (i *Instance) getPackage(r *http.Request) any {
+	// TODO(adonovan): shouldn't we report an HTTP error in all
+	// these early returns? Same for getMetadata.
+	session := i.State.Session(r.PathValue("session"))
+	if session == nil {
+		return nil // not found
+	}
+
+	v, err := session.View(r.PathValue("view"))
+	if err != nil {
+		stdlog.Printf("/package: %v", err)
+		return nil // not found
+	}
+
+	id := r.PathValue("id")
+
+	snapshot, release, err := v.Snapshot()
+	if err != nil {
+		stdlog.Printf("/package: failed to get latest snapshot: %v", err)
+		return nil
+	}
+	defer release()
+
+	pkgs, err := snapshot.TypeCheck(r.Context(), cache.PackageID(id))
+	if err != nil {
+		stdlog.Printf("/package: failed to typecheck package %q: %v", id, err)
+		return nil
+	}
+
+	// (PackageDiagnostics is redundant w.r.t. TypeCheck but it's
+	// the only way to access type errors in cache.Diagnostic form.)
+	diags, err := snapshot.PackageDiagnostics(r.Context(), cache.PackageID(id))
+	if err != nil {
+		stdlog.Printf("/package: failed to typecheck package %q: %v", id, err)
+		return nil
+	}
+
+	return &PackageInfo{
+		SessionID:   session.ID(),
+		ViewID:      v.ID(),
+		Package:     pkgs[0],
+		Diagnostics: diags,
+	}
+}
+
+func (i *Instance) getInfo(r *http.Request) any {
 	buf := &bytes.Buffer{}
-	i.PrintServerInfo(r.Context(), buf)
+	i.writeServerInfo(buf)
 	return template.HTML(buf.String())
 }
 
@@ -340,7 +432,7 @@ func (i *Instance) AddService(s protocol.Server, session *cache.Session) {
 	stdlog.Printf("unable to find a Client to add the protocol.Server to")
 }
 
-func getMemory(_ *http.Request) interface{} {
+func getMemory(_ *http.Request) any {
 	var m runtime.MemStats
 	runtime.ReadMemStats(&m)
 	return m
@@ -363,20 +455,21 @@ func GetInstance(ctx context.Context) *Instance {
 
 // WithInstance creates debug instance ready for use using the supplied
 // configuration and stores it in the returned context.
-func WithInstance(ctx context.Context, agent string) context.Context {
+func WithInstance(ctx context.Context, otelEndpoint string) context.Context {
 	i := &Instance{
-		StartTime:     time.Now(),
-		OCAgentConfig: agent,
+		StartTime: time.Now(),
 	}
 	i.LogWriter = os.Stderr
-	ocConfig := ocagent.Discover()
-	//TODO: we should not need to adjust the discovered configuration
-	ocConfig.Address = i.OCAgentConfig
-	i.ocagent = ocagent.Connect(ocConfig)
 	i.prometheus = prometheus.New()
 	i.rpcs = &Rpcs{}
 	i.traces = &traces{}
 	i.State = &State{}
+	if otelEndpoint != "" {
+		i.otel = otel.NewExporter(ctx,
+			otel.WithEndpoint(otelEndpoint),
+			otel.WithServiceName("gopls"),
+			otel.WithServiceVersion(versionpkg.Version()))
+	}
 	i.exporter = makeInstanceExporter(i)
 	return context.WithValue(ctx, instanceKey, i)
 }
@@ -439,13 +532,20 @@ func (i *Instance) Serve(ctx context.Context, addr string) (string, error) {
 	event.Log(ctx, "Debug serving", label1.Port.Of(port))
 	go func() {
 		mux := http.NewServeMux()
-		mux.HandleFunc("/", render(MainTmpl, func(*http.Request) interface{} { return i }))
+		mux.HandleFunc("/", render(MainTmpl, func(*http.Request) any { return i }))
 		mux.HandleFunc("/debug/", render(DebugTmpl, nil))
 		mux.HandleFunc("/debug/pprof/", pprof.Index)
 		mux.HandleFunc("/debug/pprof/cmdline", cmdline)
 		mux.HandleFunc("/debug/pprof/profile", pprof.Profile)
 		mux.HandleFunc("/debug/pprof/symbol", pprof.Symbol)
 		mux.HandleFunc("/debug/pprof/trace", pprof.Trace)
+
+		if h, err := startFlightRecorder(); err != nil {
+			stdlog.Printf("failed to start flight recorder: %v", err) // e.g. go1.24
+		} else {
+			mux.HandleFunc("/flightrecorder", h)
+		}
+
 		if i.prometheus != nil {
 			mux.HandleFunc("/metrics/", i.prometheus.Serve)
 		}
@@ -460,7 +560,9 @@ func (i *Instance) Serve(ctx context.Context, addr string) (string, error) {
 		mux.HandleFunc("/session/", render(SessionTmpl, i.getSession))
 		mux.HandleFunc("/client/", render(ClientTmpl, i.getClient))
 		mux.HandleFunc("/server/", render(ServerTmpl, i.getServer))
-		mux.HandleFunc("/file/", render(FileTmpl, i.getFile))
+		mux.HandleFunc("/file/{session}/{identifier}", render(FileTmpl, i.getFile))
+		mux.HandleFunc("/metadata/{session}/{view}/", render(MetadataTmpl, i.getMetadata))
+		mux.HandleFunc("/package/{session}/{view}/{id...}", render(PackageTmpl, i.getPackage))
 		mux.HandleFunc("/info", render(InfoTmpl, i.getInfo))
 		mux.HandleFunc("/memory", render(MemoryTmpl, getMemory))
 
@@ -476,11 +578,8 @@ func (i *Instance) Serve(ctx context.Context, addr string) (string, error) {
 			http.Error(w, "made a bug", http.StatusOK)
 		})
 
-		if err := http.Serve(listener, mux); err != nil {
-			event.Error(ctx, "Debug server failed", err)
-			return
-		}
-		event.Log(ctx, "Debug server finished")
+		err := http.Serve(listener, mux) // always non-nil
+		event.Error(ctx, "Debug server failed", err)
 	}()
 	return i.listenedDebugAddress, nil
 }
@@ -498,8 +597,10 @@ func (i *Instance) ListenedDebugAddress() string {
 }
 
 func makeGlobalExporter(stderr io.Writer) event.Exporter {
-	p := export.Printer{}
-	var pMu sync.Mutex
+	var (
+		pMu sync.Mutex
+		p   export.Printer
+	)
 	return func(ctx context.Context, ev core.Event, lm label.Map) context.Context {
 		i := GetInstance(ctx)
 
@@ -541,9 +642,6 @@ func messageType(l log.Level) protocol.MessageType {
 
 func makeInstanceExporter(i *Instance) event.Exporter {
 	exporter := func(ctx context.Context, ev core.Event, lm label.Map) context.Context {
-		if i.ocagent != nil {
-			ctx = i.ocagent.ProcessEvent(ctx, ev, lm)
-		}
 		if i.prometheus != nil {
 			ctx = i.prometheus.ProcessEvent(ctx, ev, lm)
 		}
@@ -553,6 +651,9 @@ func makeInstanceExporter(i *Instance) event.Exporter {
 		if i.traces != nil {
 			ctx = i.traces.ProcessEvent(ctx, ev, lm)
 		}
+		if i.otel != nil {
+			ctx = i.otel.ProcessEvent(ctx, ev, lm)
+		}
 		if event.IsLog(ev) {
 			if s := cache.KeyCreateSession.Get(ev); s != nil {
 				i.State.addClient(s)
@@ -594,11 +695,11 @@ func makeInstanceExporter(i *Instance) event.Exporter {
 	return exporter
 }
 
-type dataFunc func(*http.Request) interface{}
+type dataFunc func(*http.Request) any
 
 func render(tmpl *template.Template, fun dataFunc) func(http.ResponseWriter, *http.Request) {
 	return func(w http.ResponseWriter, r *http.Request) {
-		var data interface{}
+		var data any
 		if fun != nil {
 			data = fun(r)
 		}
@@ -661,6 +762,7 @@ body {
 <a href="/metrics">Metrics</a>
 <a href="/rpc">RPC</a>
 <a href="/trace">Trace</a>
+<a href="/flightrecorder">Flight recorder</a>
 <a href="/analysis">Analysis</a>
 <hr>
 <h1>{{template "title" .}}</h1>
@@ -811,6 +913,8 @@ var SessionTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
 {{define "title"}}Session {{.ID}}{{end}}
 {{define "body"}}
 From: <b>{{template "cachelink" .Cache.ID}}</b><br>
+{{- $session := . -}}
+
 <h2>Views</h2>
 <ul>{{range .Views}}
 {{- $envOverlay := .EnvOverlay -}}
@@ -820,8 +924,17 @@ Root: <b>{{.Root}}</b><br>
 {{- if $envOverlay}}
 Env overlay: <b>{{$envOverlay}})</b><br>
 {{end -}}
-Folder: <b>{{.Folder.Name}}:{{.Folder.Dir}}</b></li>
+Folder.Name: <b>{{.Folder.Name}}</b><br>
+Folder.Dir: <b>{{.Folder.Dir}}</b><br/>
+<a href="/metadata/{{$session.ID}}/{{.ID}}">Latest metadata</a><br/>
+
+Settings:<br/>
+<ul>
+{{range .Folder.Options.Debug}}<li>{{.}}</li>
+{{end}}
+</ul>
 {{end}}</ul>
+
 <h2>Overlays</h2>
 {{$session := .}}
 <ul>{{range .Overlays}}
@@ -831,6 +944,7 @@ Folder: <b>{{.Folder.Name}}:{{.Folder.Dir}}</b></li>
 {{end}}
 `))
 
+// For /file endpoint; operand is [FileWithKind].
 var FileTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
 {{define "title"}}Overlay {{.Identity.Hash}}{{end}}
 {{define "body"}}
@@ -844,3 +958,91 @@ var FileTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
 <pre>{{fcontent .Content}}</pre>
 {{end}}
 `))
+
+// For /metadata endpoint; operand is [*MetadataInfo].
+var MetadataTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
+{{define "title"}}Metadata graph{{end}}
+{{define "body"}}
+
+<p><a href='#hdr-Files'>↓ Index by file</a></p>
+
+<h3>Packages ({{len .Graph.Packages}})</h3>
+<ul>
+{{range $id, $pkg := .Graph.Packages}}
+<li id='{{$id}}'><b>{{$id}}</b>
+{{with $pkg}}
+<ul>
+ <li>Name: {{.Name}}</li>
+ <li>PkgPath: {{printf "%q" .PkgPath}}</li>
+ {{if .Module}}<li>Module: {{printf "%#v" .Module}}</li>{{end}}
+ <li><a href="/package/{{$.SessionID}}/{{$.ViewID}}/{{$id}}">Type information</a></li>
+ {{if .ForTest}}<li>ForTest: {{.ForTest}}</li>{{end}}
+ {{if .Standalone}}<li>Standalone</li>{{end}}
+ {{if .Errors}}<li>Errors: {{.Errors}}</li>{{end}}
+ {{if .DepsErrors}}<li>DepsErrors: {{.DepsErrors}}</li>{{end}}
+ <li>LoadDir: {{.LoadDir}}</li>
+ <li>DepsByImpPath
+  <ul>
+   {{range $path, $id := .DepsByImpPath}}
+    <li>{{if $id}}<a href='#{{$id}}'>{{printf "%q" $path}}</a>{{else}}⚠️ {{printf "%q" $path}} missing{{end}}</li>
+   {{end}}
+  </ul>
+ </li>
+ {{if .GoFiles}}<li>GoFiles: <ul>{{range .GoFiles}}<li>{{.}}</li>{{end}}</ul></li>{{end}}
+ {{if .CompiledGoFiles}}<li>CompiledGoFiles: <ul>{{range .CompiledGoFiles}}<li>{{.}}</li>{{end}}</ul></li>{{end}}
+ {{if .IgnoredFiles}}<li>IgnoredFiles: <ul>{{range .IgnoredFiles}}<li>{{.}}</li>{{end}}</ul></li>{{end}}
+ {{if .OtherFiles}}<li>OtherFiles: <ul>{{range .OtherFiles}}<li>{{.}}</li>{{end}}</ul></li>{{end}}
+ <!-- skip DepsByPkgPath, ImportedBy (redundant indexes) -->
+</ul>
+{{end}}
+</li>
+{{end}}
+</ul>
+
+<h3 id='hdr-Files'>Files</h3>
+<ul>
+{{range $uri, $pkgs := .Graph.ForFile}}<li>{{$uri}} →{{range $pkgs}} <a href='#{{.ID}}'>{{.ID}}</a>{{end}}</li>{{end}}
+</ul>
+
+{{end}}
+`))
+
+// For /package endpoint; operand is [*PackageInfo].
+var PackageTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
+{{define "title"}}Package {{.Package.Metadata.ID}}{{end}}
+{{define "body"}}
+
+<ul>
+<li><a href="/metadata/{{.SessionID}}/{{.ViewID}}#{{.Package.Metadata.ID}}">Metadata</a></li>
+</ul>
+
+<h2>Diagnostics for syntax and type (but not analysis) errors</h2>
+<ul>
+ {{range $url, $diags := .Diagnostics}}
+ <li>{{$url}}
+  <ul>
+  {{range $diag := $diags}}
+  <li>{{$diag.Range}}: [{{$diag.Severity}}] {{$diag.Message}}<br/>
+  <ul>
+    <li>code {{$diag.Code}}</li>
+    <li>code href {{$diag.CodeHref}}</li>
+    <li>source {{$diag.Source}}</li>
+    <li>tags {{$diag.Tags}}</li>
+    <li>related {{$diag.Related}}</li> {{/*TODO: improve*/}}
+    <li>bundled fixes {{$diag.BundledFixes}}</li> {{/*TODO: improve*/}}
+    <li>fixes {{$diag.SuggestedFixes}}</li> {{/*TODO: improve*/}}
+  </ul>
+  {{end}}
+ </ul>
+ {{end}}
+</ul>
+
+{{end}}
+{{/*
+TODO:
+ - link to godoc (tricky: in server package)
+ - show Object inventory of types.Package.Scope
+ - show index info (xrefs, methodsets, tests)
+ - call DiagnoseFile on each file?
+*/}}
+`))
