@@ -10,16 +10,22 @@ import (
 	"go/ast"
 	"go/token"
 	"go/types"
+	"strconv"
+	"strings"
 
-	"golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/edge"
+	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/protocol"
-	"golang.org/x/tools/gopls/internal/util/typesutil"
+	"golang.org/x/tools/gopls/internal/util/cursorutil"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/event"
+	"golang.org/x/tools/internal/fmtstr"
+	"golang.org/x/tools/internal/moreiters"
 )
 
-func Highlight(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, position protocol.Position) ([]protocol.DocumentHighlight, error) {
+func Highlight(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, rng protocol.Range) ([]protocol.DocumentHighlight, error) {
 	ctx, done := event.Start(ctx, "golang.Highlight")
 	defer done()
 
@@ -30,33 +36,20 @@ func Highlight(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, po
 		return nil, fmt.Errorf("getting package for Highlight: %w", err)
 	}
 
-	pos, err := pgf.PositionPos(position)
+	start, end, err := pgf.RangePos(rng)
 	if err != nil {
 		return nil, err
 	}
-	path, _ := astutil.PathEnclosingInterval(pgf.File, pos, pos)
-	if len(path) == 0 {
-		return nil, fmt.Errorf("no enclosing position found for %v:%v", position.Line, position.Character)
-	}
-	// If start == end for astutil.PathEnclosingInterval, the 1-char interval
-	// following start is used instead. As a result, we might not get an exact
-	// match so we should check the 1-char interval to the left of the passed
-	// in position to see if that is an exact match.
-	if _, ok := path[0].(*ast.Ident); !ok {
-		if p, _ := astutil.PathEnclosingInterval(pgf.File, pos-1, pos-1); p != nil {
-			switch p[0].(type) {
-			case *ast.Ident, *ast.SelectorExpr:
-				path = p // use preceding ident/selector
-			}
-		}
-	}
-	result, err := highlightPath(path, pgf.File, pkg.TypesInfo())
+
+	cur, _, _, _ := astutil.Select(pgf.Cursor(), start, end) // can't fail: pgf contains pos
+
+	result, err := highlightPath(pkg.TypesInfo(), cur, start, end)
 	if err != nil {
 		return nil, err
 	}
 	var ranges []protocol.DocumentHighlight
 	for rng, kind := range result {
-		rng, err := pgf.PosRange(rng.start, rng.end)
+		rng, err := pgf.NodeRange(rng)
 		if err != nil {
 			return nil, err
 		}
@@ -68,46 +61,62 @@ func Highlight(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, po
 	return ranges, nil
 }
 
-// highlightPath returns ranges to highlight for the given enclosing path,
-// which should be the result of astutil.PathEnclosingInterval.
-func highlightPath(path []ast.Node, file *ast.File, info *types.Info) (map[posRange]protocol.DocumentHighlightKind, error) {
-	result := make(map[posRange]protocol.DocumentHighlightKind)
-	switch node := path[0].(type) {
+// highlightPath returns ranges to highlight for the given the cursor.
+func highlightPath(info *types.Info, cur inspector.Cursor, start, end token.Pos) (map[astutil.Range]protocol.DocumentHighlightKind, error) {
+	result := make(map[astutil.Range]protocol.DocumentHighlightKind)
+
+	// Inside a call to a printf-like function (as identified
+	// by a simple heuristic).
+	// Treat each corresponding ("%v", arg) pair as a highlight class.
+	for node := range cur.Enclosing((*ast.CallExpr)(nil)) {
+		call := node.Node().(*ast.CallExpr)
+		lit, idx := formatStringAndIndex(info, call)
+		if idx != -1 {
+			highlightPrintf(call, idx, start, end, lit, result)
+		}
+	}
+
+	switch node := cur.Node().(type) {
 	case *ast.BasicLit:
 		// Import path string literal?
-		if len(path) > 1 {
-			if imp, ok := path[1].(*ast.ImportSpec); ok {
-				highlight := func(n ast.Node) {
-					highlightNode(result, n, protocol.Text)
-				}
-
-				// Highlight the import itself...
-				highlight(imp)
-
-				// ...and all references to it in the file.
-				if pkgname, ok := typesutil.ImportedPkgName(info, imp); ok {
-					ast.Inspect(file, func(n ast.Node) bool {
-						if id, ok := n.(*ast.Ident); ok &&
-							info.Uses[id] == pkgname {
-							highlight(id)
-						}
-						return true
-					})
-				}
-				return result, nil
+		if imp, ok := cur.Parent().Node().(*ast.ImportSpec); ok {
+			highlight := func(n ast.Node) {
+				highlightNode(result, n, protocol.Text)
 			}
+
+			// Highlight the import itself...
+			highlight(imp)
+
+			// ...and all references to it in the file.
+			if pkgname := info.PkgNameOf(imp); pkgname != nil {
+				_, curFile := cursorutil.FirstEnclosing[*ast.File](cur)
+				for c := range curFile.Preorder((*ast.Ident)(nil)) {
+					if id := c.Node().(*ast.Ident); info.Uses[id] == pkgname {
+						highlight(id)
+					}
+				}
+			}
+			return result, nil
 		}
-		highlightFuncControlFlow(path, result)
+		highlightFuncControlFlow(cur, result)
 	case *ast.ReturnStmt, *ast.FuncDecl, *ast.FuncType:
-		highlightFuncControlFlow(path, result)
+		highlightFuncControlFlow(cur, result)
 	case *ast.Ident:
 		// Check if ident is inside return or func decl.
-		highlightFuncControlFlow(path, result)
-		highlightIdentifier(node, file, info, result)
+		highlightFuncControlFlow(cur, result)
+		highlightIdentifier(cur, info, result)
 	case *ast.ForStmt, *ast.RangeStmt:
-		highlightLoopControlFlow(path, info, result)
+		var label *ast.Ident
+		if l, ok := cur.Parent().Node().(*ast.LabeledStmt); ok {
+			label = l.Label
+		}
+		highlightLoopControlFlow(node, label, info, result)
 	case *ast.SwitchStmt, *ast.TypeSwitchStmt:
-		highlightSwitchFlow(path, info, result)
+		var label *ast.Ident
+		if l, ok := cur.Parent().Node().(*ast.LabeledStmt); ok {
+			label = l.Label
+		}
+		highlightSwitchFlow(node, label, info, result)
 	case *ast.BranchStmt:
 		// BREAK can exit a loop, switch or select, while CONTINUE exit a loop so
 		// these need to be handled separately. They can also be embedded in any
@@ -116,15 +125,25 @@ func highlightPath(path []ast.Node, file *ast.File, info *types.Info) (map[posRa
 		switch node.Tok {
 		case token.BREAK:
 			if node.Label != nil {
-				highlightLabeledFlow(path, info, node, result)
+				highlightLabeledFlow(cur, info, result)
 			} else {
-				highlightUnlabeledBreakFlow(path, info, result)
+				highlightUnlabeledBreakFlow(cur, info, result)
 			}
 		case token.CONTINUE:
 			if node.Label != nil {
-				highlightLabeledFlow(path, info, node, result)
+				highlightLabeledFlow(cur, info, result)
 			} else {
-				highlightLoopControlFlow(path, info, result)
+				var (
+					stmt  ast.Node
+					label *ast.Ident
+				)
+				if curLoop, ok := moreiters.First(cur.Enclosing((*ast.ForStmt)(nil), (*ast.RangeStmt)(nil))); ok {
+					stmt = curLoop.Node()
+					if l, ok := curLoop.Parent().Node().(*ast.LabeledStmt); ok {
+						label = l.Label
+					}
+				}
+				highlightLoopControlFlow(stmt, label, info, result)
 			}
 		}
 	}
@@ -132,8 +151,142 @@ func highlightPath(path []ast.Node, file *ast.File, info *types.Info) (map[posRa
 	return result, nil
 }
 
-type posRange struct {
-	start, end token.Pos
+// formatStringAndIndex returns the BasicLit and index of the BasicLit (the last
+// non-variadic parameter) within the given printf-like call
+// expression, returns -1 as index if unknown.
+func formatStringAndIndex(info *types.Info, call *ast.CallExpr) (*ast.BasicLit, int) {
+	typ := info.Types[call.Fun].Type
+	if typ == nil {
+		return nil, -1 // missing type
+	}
+	sig, ok := typ.(*types.Signature)
+	if !ok {
+		return nil, -1 // ill-typed
+	}
+	if !sig.Variadic() {
+		// Skip checking non-variadic functions.
+		return nil, -1
+	}
+	idx := sig.Params().Len() - 2
+	if !(0 <= idx && idx < len(call.Args)) {
+		// Skip checking functions without a format string parameter, or
+		// missing the corresponding format argument.
+		return nil, -1
+	}
+	// We only care about literal format strings, so fmt.Sprint("a"+"b%s", "bar") won't be highlighted.
+	if lit, ok := call.Args[idx].(*ast.BasicLit); ok && lit.Kind == token.STRING {
+		return lit, idx
+	}
+	return nil, -1
+}
+
+// highlightPrintf highlights operations in a format string and their
+// corresponding variadic arguments in a (possible) printf-style function call.
+// For example:
+//
+// fmt.Printf("Hello %s, you scored %d", name, score)
+//
+// If the cursor is on %s or name, it will highlight %s as a write operation,
+// and name as a read operation.
+func highlightPrintf(call *ast.CallExpr, idx int, start, end token.Pos, lit *ast.BasicLit, result map[astutil.Range]protocol.DocumentHighlightKind) {
+	format, err := strconv.Unquote(lit.Value)
+	if err != nil {
+		return
+	}
+	if !strings.Contains(format, "%") {
+		return
+	}
+	operations, err := fmtstr.Parse(format, idx)
+	if err != nil {
+		return
+	}
+
+	// fmt.Printf("%[1]d %[1].2d", 3)
+	//
+	// When cursor is in `%[1]d`, we record `3` being successfully highlighted.
+	// And because we will also record `%[1].2d`'s corresponding arguments index is `3`
+	// in `visited`, even though it will not highlight any item in the first pass,
+	// in the second pass we can correctly highlight it. So the three are the same class.
+	succeededArg := 0
+	visited := make(map[astutil.Range]int, 0)
+
+	// highlightPair highlights the operation and its potential argument pair if the cursor is within either range.
+	highlightPair := func(rang fmtstr.Range, argIndex int) {
+		var (
+			rng       astutil.Range
+			withinRng bool
+		)
+		rng, err = astutil.RangeInStringLiteral(lit, rang.Start, rang.End)
+		if err != nil {
+			return
+		}
+		visited[rng] = argIndex
+
+		if start == end {
+			// End pos can't equal to range's end, otherwise the two neighborhood
+			// such as (%[2]*d) are both highlighted if cursor in "d" (ending of [2]*).
+			withinRng = rng.Start <= start && end < rng.End()
+		} else {
+			// With a non-empty selection, there is no ambiguity.
+			withinRng = rng.Contains(astutil.RangeOf(start, end))
+		}
+
+		var (
+			arg       ast.Expr
+			withinArg bool
+		)
+		if argIndex < len(call.Args) {
+			arg = call.Args[argIndex]
+			withinArg = astutil.NodeContains(arg, astutil.RangeOf(start, end))
+		}
+
+		if withinRng || withinArg {
+			highlightRange(result, rng, protocol.Write)
+			if arg != nil {
+				succeededArg = argIndex
+				highlightNode(result, arg, protocol.Read)
+			}
+		}
+	}
+
+	for _, op := range operations {
+		// If width or prec has any *, we can not highlight the full range from % to verb,
+		// because it will overlap with the sub-range of *, for example:
+		//
+		// fmt.Printf("%*[3]d", 4, 5, 6)
+		//               ^  ^ we can only highlight this range when cursor in 6. '*' as a one-rune range will
+		//               highlight for 4.
+		hasAsterisk := false
+
+		// Try highlight Width if there is a *.
+		if op.Width.Dynamic != -1 {
+			hasAsterisk = true
+			highlightPair(op.Width.Range, op.Width.Dynamic)
+		}
+
+		// Try highlight Precision if there is a *.
+		if op.Prec.Dynamic != -1 {
+			hasAsterisk = true
+			highlightPair(op.Prec.Range, op.Prec.Dynamic)
+		}
+
+		// Try highlight Verb.
+		if op.Verb.Verb != '%' {
+			// If any * is found inside operation, narrow the highlight range.
+			if hasAsterisk {
+				highlightPair(op.Verb.Range, op.Verb.ArgIndex)
+			} else {
+				highlightPair(op.Range, op.Verb.ArgIndex)
+			}
+		}
+	}
+
+	// Second pass, try to highlight those missed operations.
+	for rng, argIndex := range visited {
+		if succeededArg == argIndex {
+			highlightRange(result, rng, protocol.Write)
+		}
+	}
 }
 
 // highlightFuncControlFlow adds highlight ranges to the result map to
@@ -146,7 +299,7 @@ type posRange struct {
 //
 // As a special case, if the cursor is within a complicated expression, control
 // flow highlighting is disabled, as it would highlight too much.
-func highlightFuncControlFlow(path []ast.Node, result map[posRange]protocol.DocumentHighlightKind) {
+func highlightFuncControlFlow(cur inspector.Cursor, result map[astutil.Range]protocol.DocumentHighlightKind) {
 
 	var (
 		funcType   *ast.FuncType   // type of enclosing func, or nil
@@ -154,9 +307,9 @@ func highlightFuncControlFlow(path []ast.Node, result map[posRange]protocol.Docu
 		returnStmt *ast.ReturnStmt // enclosing ReturnStmt within the func, or nil
 	)
 
-findEnclosingFunc:
-	for i, n := range path {
-		switch n := n.(type) {
+loop:
+	for cur := range cur.Enclosing() {
+		switch n := cur.Node().(type) {
 		// TODO(rfindley, low priority): these pre-existing cases for KeyValueExpr
 		// and CallExpr appear to avoid highlighting when the cursor is in a
 		// complicated expression. However, the basis for this heuristic is
@@ -165,29 +318,24 @@ findEnclosingFunc:
 			// If cursor is in a key: value expr, we don't want control flow highlighting.
 			return
 
-		case *ast.CallExpr:
-			// If cursor is an arg in a callExpr, we don't want control flow highlighting.
-			if i > 0 {
-				for _, arg := range n.Args {
-					if arg == path[i-1] {
-						return
-					}
-				}
-			}
-
 		case *ast.FuncLit:
 			funcType = n.Type
 			funcBody = n.Body
-			break findEnclosingFunc
+			break loop
 
 		case *ast.FuncDecl:
 			funcType = n.Type
 			funcBody = n.Body
-			break findEnclosingFunc
+			break loop
 
 		case *ast.ReturnStmt:
 			returnStmt = n
 		}
+
+		// If cursor is an arg in a callExpr, we don't want control flow highlighting.
+		if cur.ParentEdgeKind() == edge.CallExpr_Args {
+			return
+		}
 	}
 
 	if funcType == nil {
@@ -196,7 +344,7 @@ findEnclosingFunc:
 
 	// Helper functions for inspecting the current location.
 	var (
-		pos    = path[0].Pos()
+		pos    = cur.Node().Pos()
 		inSpan = func(start, end token.Pos) bool { return start <= pos && pos < end }
 		inNode = func(n ast.Node) bool { return inSpan(n.Pos(), n.End()) }
 	)
@@ -207,16 +355,16 @@ findEnclosingFunc:
 	// specific field or expression, we should highlight all of the exit points
 	// of the function, including the "return" and "func" keywords.
 	funcEnd := funcType.Func + token.Pos(len("func"))
-	highlightAll := path[0] == returnStmt || inSpan(funcType.Func, funcEnd)
+	highlightAll := cur.Node() == returnStmt || inSpan(funcType.Func, funcEnd)
 	var highlightIndexes map[int]bool
 
 	if highlightAll {
 		// Add the "func" part of the func declaration.
-		highlightRange(result, funcType.Func, funcEnd, protocol.Text)
+		highlightRange(result, astutil.RangeOf(funcType.Func, funcEnd), protocol.Text)
 	} else if returnStmt == nil && !inResults {
 		return // nothing to highlight
 	} else {
-		// If we're not highighting the entire return statement, we need to collect
+		// If we're not highlighting the entire return statement, we need to collect
 		// specific result indexes to highlight. This may be more than one index if
 		// the cursor is on a multi-name result field, but not in any specific name.
 		if !highlightAll {
@@ -301,16 +449,25 @@ findEnclosingFunc:
 	}
 }
 
-// highlightUnlabeledBreakFlow highlights the innermost enclosing for/range/switch or swlect
-func highlightUnlabeledBreakFlow(path []ast.Node, info *types.Info, result map[posRange]protocol.DocumentHighlightKind) {
+// highlightUnlabeledBreakFlow highlights the innermost enclosing
+// for/range/switch or select
+func highlightUnlabeledBreakFlow(cur inspector.Cursor, info *types.Info, result map[astutil.Range]protocol.DocumentHighlightKind) {
 	// Reverse walk the path until we find closest loop, select, or switch.
-	for _, n := range path {
-		switch n.(type) {
+	for cur := range cur.Enclosing() {
+		switch n := cur.Node().(type) {
 		case *ast.ForStmt, *ast.RangeStmt:
-			highlightLoopControlFlow(path, info, result)
+			var label *ast.Ident
+			if l, ok := cur.Parent().Node().(*ast.LabeledStmt); ok {
+				label = l.Label
+			}
+			highlightLoopControlFlow(n, label, info, result)
 			return // only highlight the innermost statement
 		case *ast.SwitchStmt, *ast.TypeSwitchStmt:
-			highlightSwitchFlow(path, info, result)
+			var label *ast.Ident
+			if l, ok := cur.Parent().Node().(*ast.LabeledStmt); ok {
+				label = l.Label
+			}
+			highlightSwitchFlow(n, label, info, result)
 			return
 		case *ast.SelectStmt:
 			// TODO: add highlight when breaking a select.
@@ -321,58 +478,31 @@ func highlightUnlabeledBreakFlow(path []ast.Node, info *types.Info, result map[p
 
 // highlightLabeledFlow highlights the enclosing labeled for, range,
 // or switch statement denoted by a labeled break or continue stmt.
-func highlightLabeledFlow(path []ast.Node, info *types.Info, stmt *ast.BranchStmt, result map[posRange]protocol.DocumentHighlightKind) {
-	use := info.Uses[stmt.Label]
+//
+// The input cursor must point to a [ast.BranchStmt]
+func highlightLabeledFlow(curBranch inspector.Cursor, info *types.Info, result map[astutil.Range]protocol.DocumentHighlightKind) {
+	use := info.Uses[curBranch.Node().(*ast.BranchStmt).Label]
 	if use == nil {
 		return
 	}
-	for _, n := range path {
-		if label, ok := n.(*ast.LabeledStmt); ok && info.Defs[label.Label] == use {
+	for cur := range curBranch.Enclosing((*ast.LabeledStmt)(nil)) {
+		if label := cur.Node().(*ast.LabeledStmt); info.Defs[label.Label] == use {
 			switch label.Stmt.(type) {
 			case *ast.ForStmt, *ast.RangeStmt:
-				highlightLoopControlFlow([]ast.Node{label.Stmt, label}, info, result)
+				highlightLoopControlFlow(label.Stmt, label.Label, info, result)
 			case *ast.SwitchStmt, *ast.TypeSwitchStmt:
-				highlightSwitchFlow([]ast.Node{label.Stmt, label}, info, result)
+				highlightSwitchFlow(label.Stmt, label.Label, info, result)
 			}
 			return
 		}
 	}
 }
 
-func labelFor(path []ast.Node) *ast.Ident {
-	if len(path) > 1 {
-		if n, ok := path[1].(*ast.LabeledStmt); ok {
-			return n.Label
-		}
-	}
-	return nil
-}
-
-func highlightLoopControlFlow(path []ast.Node, info *types.Info, result map[posRange]protocol.DocumentHighlightKind) {
-	var loop ast.Node
-	var loopLabel *ast.Ident
-	stmtLabel := labelFor(path)
-Outer:
-	// Reverse walk the path till we get to the for loop.
-	for i := range path {
-		switch n := path[i].(type) {
-		case *ast.ForStmt, *ast.RangeStmt:
-			loopLabel = labelFor(path[i:])
-
-			if stmtLabel == nil || loopLabel == stmtLabel {
-				loop = n
-				break Outer
-			}
-		}
-	}
-	if loop == nil {
-		return
-	}
-
+func highlightLoopControlFlow(loop ast.Node, label *ast.Ident, info *types.Info, result map[astutil.Range]protocol.DocumentHighlightKind) {
 	// Add the for statement.
 	rngStart := loop.Pos()
 	rngEnd := loop.Pos() + token.Pos(len("for"))
-	highlightRange(result, rngStart, rngEnd, protocol.Text)
+	highlightRange(result, astutil.RangeOf(rngStart, rngEnd), protocol.Text)
 
 	// Traverse AST to find branch statements within the same for-loop.
 	ast.Inspect(loop, func(n ast.Node) bool {
@@ -386,7 +516,7 @@ Outer:
 		if !ok {
 			return true
 		}
-		if b.Label == nil || info.Uses[b.Label] == info.Defs[loopLabel] {
+		if b.Label == nil || info.Uses[b.Label] == info.Defs[label] {
 			highlightNode(result, b, protocol.Text)
 		}
 		return true
@@ -406,7 +536,7 @@ Outer:
 	})
 
 	// We don't need to check other for loops if we aren't looking for labeled statements.
-	if loopLabel == nil {
+	if label == nil {
 		return
 	}
 
@@ -417,44 +547,24 @@ Outer:
 			return true
 		}
 		// statement with labels that matches the loop
-		if b.Label != nil && info.Uses[b.Label] == info.Defs[loopLabel] {
+		if b.Label != nil && info.Uses[b.Label] == info.Defs[label] {
 			highlightNode(result, b, protocol.Text)
 		}
 		return true
 	})
 }
 
-func highlightSwitchFlow(path []ast.Node, info *types.Info, result map[posRange]protocol.DocumentHighlightKind) {
-	var switchNode ast.Node
-	var switchNodeLabel *ast.Ident
-	stmtLabel := labelFor(path)
-Outer:
-	// Reverse walk the path till we get to the switch statement.
-	for i := range path {
-		switch n := path[i].(type) {
-		case *ast.SwitchStmt, *ast.TypeSwitchStmt:
-			switchNodeLabel = labelFor(path[i:])
-			if stmtLabel == nil || switchNodeLabel == stmtLabel {
-				switchNode = n
-				break Outer
-			}
-		}
-	}
-	// Cursor is not in a switch statement
-	if switchNode == nil {
-		return
-	}
-
+func highlightSwitchFlow(node ast.Node, label *ast.Ident, info *types.Info, result map[astutil.Range]protocol.DocumentHighlightKind) {
 	// Add the switch statement.
-	rngStart := switchNode.Pos()
-	rngEnd := switchNode.Pos() + token.Pos(len("switch"))
-	highlightRange(result, rngStart, rngEnd, protocol.Text)
+	rngStart := node.Pos()
+	rngEnd := node.Pos() + token.Pos(len("switch"))
+	highlightRange(result, astutil.RangeOf(rngStart, rngEnd), protocol.Text)
 
 	// Traverse AST to find break statements within the same switch.
-	ast.Inspect(switchNode, func(n ast.Node) bool {
+	ast.Inspect(node, func(n ast.Node) bool {
 		switch n.(type) {
 		case *ast.SwitchStmt, *ast.TypeSwitchStmt:
-			return switchNode == n
+			return node == n
 		case *ast.ForStmt, *ast.RangeStmt, *ast.SelectStmt:
 			return false
 		}
@@ -464,25 +574,25 @@ Outer:
 			return true
 		}
 
-		if b.Label == nil || info.Uses[b.Label] == info.Defs[switchNodeLabel] {
+		if b.Label == nil || info.Uses[b.Label] == info.Defs[label] {
 			highlightNode(result, b, protocol.Text)
 		}
 		return true
 	})
 
 	// We don't need to check other switches if we aren't looking for labeled statements.
-	if switchNodeLabel == nil {
+	if label == nil {
 		return
 	}
 
 	// Find labeled break statements in any switch
-	ast.Inspect(switchNode, func(n ast.Node) bool {
+	ast.Inspect(node, func(n ast.Node) bool {
 		b, ok := n.(*ast.BranchStmt)
 		if !ok || b.Tok != token.BREAK {
 			return true
 		}
 
-		if b.Label != nil && info.Uses[b.Label] == info.Defs[switchNodeLabel] {
+		if b.Label != nil && info.Uses[b.Label] == info.Defs[label] {
 			highlightNode(result, b, protocol.Text)
 		}
 
@@ -490,12 +600,11 @@ Outer:
 	})
 }
 
-func highlightNode(result map[posRange]protocol.DocumentHighlightKind, n ast.Node, kind protocol.DocumentHighlightKind) {
-	highlightRange(result, n.Pos(), n.End(), kind)
+func highlightNode(result map[astutil.Range]protocol.DocumentHighlightKind, n ast.Node, kind protocol.DocumentHighlightKind) {
+	highlightRange(result, astutil.NodeRange(n), kind)
 }
 
-func highlightRange(result map[posRange]protocol.DocumentHighlightKind, pos, end token.Pos, kind protocol.DocumentHighlightKind) {
-	rng := posRange{pos, end}
+func highlightRange(result map[astutil.Range]protocol.DocumentHighlightKind, rng astutil.Range, kind protocol.DocumentHighlightKind) {
 	// Order of traversal is important: some nodes (e.g. identifiers) are
 	// visited more than once, but the kind set during the first visitation "wins".
 	if _, exists := result[rng]; !exists {
@@ -503,15 +612,16 @@ func highlightRange(result map[posRange]protocol.DocumentHighlightKind, pos, end
 	}
 }
 
-func highlightIdentifier(id *ast.Ident, file *ast.File, info *types.Info, result map[posRange]protocol.DocumentHighlightKind) {
+func highlightIdentifier(cur inspector.Cursor, info *types.Info, result map[astutil.Range]protocol.DocumentHighlightKind) {
+	ident := cur.Node().(*ast.Ident)
 
 	// obj may be nil if the Ident is undefined.
 	// In this case, the behavior expected by tests is
 	// to match other undefined Idents of the same name.
-	obj := info.ObjectOf(id)
+	obj := info.ObjectOf(ident)
 
 	highlightIdent := func(n *ast.Ident, kind protocol.DocumentHighlightKind) {
-		if n.Name == id.Name && info.ObjectOf(n) == obj {
+		if n.Name == ident.Name && info.ObjectOf(n) == obj {
 			highlightNode(result, n, kind)
 		}
 	}
@@ -537,8 +647,9 @@ func highlightIdentifier(id *ast.Ident, file *ast.File, info *types.Info, result
 		}
 	}
 
-	ast.Inspect(file, func(n ast.Node) bool {
-		switch n := n.(type) {
+	_, curFile := cursorutil.FirstEnclosing[*ast.File](cur)
+	curFile.Inspect(nil, func(c inspector.Cursor) bool {
+		switch n := c.Node().(type) {
 		case *ast.AssignStmt:
 			for _, s := range n.Lhs {
 				highlightWriteInExpr(s)
@@ -559,6 +670,9 @@ func highlightIdentifier(id *ast.Ident, file *ast.File, info *types.Info, result
 			highlightWriteInExpr(n.Chan)
 		case *ast.CompositeLit:
 			t := info.TypeOf(n)
+			if t == nil {
+				t = types.Typ[types.Invalid]
+			}
 			if ptr, ok := t.Underlying().(*types.Pointer); ok {
 				t = ptr.Elem()
 			}
@@ -586,8 +700,8 @@ func highlightIdentifier(id *ast.Ident, file *ast.File, info *types.Info, result
 				highlightIdent(n, protocol.Text)
 			}
 		case *ast.ImportSpec:
-			pkgname, ok := typesutil.ImportedPkgName(info, n)
-			if ok && pkgname == obj {
+			pkgname := info.PkgNameOf(n)
+			if pkgname == obj {
 				if n.Name != nil {
 					highlightNode(result, n.Name, protocol.Text)
 				} else {
