@@ -12,20 +12,22 @@ import (
 	"bytes"
 	"go/ast"
 	"go/token"
-	"go/types"
+	"slices"
 	"sort"
 	"strings"
 
 	"golang.org/x/tools/go/analysis"
-	"golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/edge"
+	"golang.org/x/tools/go/ast/inspector"
+	"golang.org/x/tools/gopls/internal/cache"
+	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	"golang.org/x/tools/gopls/internal/util/slices"
 )
 
 // canSplitLines checks whether we can split lists of elements inside
 // an enclosing curly bracket/parens into separate lines.
-func canSplitLines(file *ast.File, fset *token.FileSet, start, end token.Pos) (string, bool, error) {
-	itemType, items, comments, _, _, _ := findSplitJoinTarget(fset, file, nil, start, end)
+func canSplitLines(curFile inspector.Cursor, fset *token.FileSet, start, end token.Pos) (string, bool, error) {
+	itemType, items, comments, _, _, _ := findSplitJoinTarget(fset, curFile, nil, start, end)
 	if itemType == "" {
 		return "", false, nil
 	}
@@ -47,8 +49,8 @@ func canSplitLines(file *ast.File, fset *token.FileSet, start, end token.Pos) (s
 
 // canJoinLines checks whether we can join lists of elements inside an
 // enclosing curly bracket/parens into a single line.
-func canJoinLines(file *ast.File, fset *token.FileSet, start, end token.Pos) (string, bool, error) {
-	itemType, items, comments, _, _, _ := findSplitJoinTarget(fset, file, nil, start, end)
+func canJoinLines(curFile inspector.Cursor, fset *token.FileSet, start, end token.Pos) (string, bool, error) {
+	itemType, items, comments, _, _, _ := findSplitJoinTarget(fset, curFile, nil, start, end)
 	if itemType == "" {
 		return "", false, nil
 	}
@@ -84,23 +86,25 @@ func canSplitJoinLines(items []ast.Node, comments []*ast.CommentGroup) bool {
 }
 
 // splitLines is a singleFile fixer.
-func splitLines(fset *token.FileSet, start, end token.Pos, src []byte, file *ast.File, _ *types.Package, _ *types.Info) (*token.FileSet, *analysis.SuggestedFix, error) {
-	itemType, items, comments, indent, braceOpen, braceClose := findSplitJoinTarget(fset, file, src, start, end)
+func splitLines(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (*token.FileSet, *analysis.SuggestedFix, error) {
+	fset := pkg.FileSet()
+	itemType, items, comments, indent, braceOpen, braceClose := findSplitJoinTarget(fset, pgf.Cursor(), pgf.Src, start, end)
 	if itemType == "" {
 		return nil, nil, nil // no fix available
 	}
 
-	return fset, processLines(fset, items, comments, src, braceOpen, braceClose, ",\n", "\n", ",\n"+indent, indent+"\t"), nil
+	return fset, processLines(fset, items, comments, pgf.Src, braceOpen, braceClose, ",\n", "\n", ",\n"+indent, indent+"\t"), nil
 }
 
 // joinLines is a singleFile fixer.
-func joinLines(fset *token.FileSet, start, end token.Pos, src []byte, file *ast.File, _ *types.Package, _ *types.Info) (*token.FileSet, *analysis.SuggestedFix, error) {
-	itemType, items, comments, _, braceOpen, braceClose := findSplitJoinTarget(fset, file, src, start, end)
+func joinLines(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (*token.FileSet, *analysis.SuggestedFix, error) {
+	fset := pkg.FileSet()
+	itemType, items, comments, _, braceOpen, braceClose := findSplitJoinTarget(fset, pgf.Cursor(), pgf.Src, start, end)
 	if itemType == "" {
 		return nil, nil, nil // no fix available
 	}
 
-	return fset, processLines(fset, items, comments, src, braceOpen, braceClose, ", ", "", "", ""), nil
+	return fset, processLines(fset, items, comments, pgf.Src, braceOpen, braceClose, ", ", "", "", ""), nil
 }
 
 // processLines is the common operation for both split and join lines because this split/join operation is
@@ -151,43 +155,49 @@ func processLines(fset *token.FileSet, items []ast.Node, comments []*ast.Comment
 		}
 
 		edits = append(edits, analysis.TextEdit{Pos: pos, End: end, NewText: []byte(sep + indent)})
+
+		// Print the Ellipsis if we synthesized one earlier.
+		if is[*ast.Ellipsis](nodes[i]) {
+			edits = append(edits, analysis.TextEdit{
+				Pos:     nodes[i].End(),
+				End:     nodes[i].End(),
+				NewText: []byte("..."),
+			})
+		}
 	}
 
 	return &analysis.SuggestedFix{TextEdits: edits}
 }
 
 // findSplitJoinTarget returns the first curly bracket/parens that encloses the current cursor.
-func findSplitJoinTarget(fset *token.FileSet, file *ast.File, src []byte, start, end token.Pos) (itemType string, items []ast.Node, comments []*ast.CommentGroup, indent string, open, close token.Pos) {
-	isCursorInside := func(nodePos, nodeEnd token.Pos) bool {
-		return nodePos < start && end < nodeEnd
-	}
+func findSplitJoinTarget(fset *token.FileSet, curFile inspector.Cursor, src []byte, start, end token.Pos) (itemType string, items []ast.Node, comments []*ast.CommentGroup, indent string, open, close token.Pos) {
 
 	findTarget := func() (targetType string, target ast.Node, open, close token.Pos) {
-		path, _ := astutil.PathEnclosingInterval(file, start, end)
-		for _, node := range path {
-			switch node := node.(type) {
-			case *ast.FuncType:
-				// params or results of func signature
-				// Note:
-				// - each ast.Field (e.g. "x, y, z int") is considered a single item.
-				// - splitting Params and Results lists is not usually good style.
-				if p := node.Params; isCursorInside(p.Opening, p.Closing) {
-					return "parameters", p, p.Opening, p.Closing
-				}
-				if r := node.Results; r != nil && isCursorInside(r.Opening, r.Closing) {
-					return "results", r, r.Opening, r.Closing
-				}
-			case *ast.CallExpr: // f(a, b, c)
-				if isCursorInside(node.Lparen, node.Rparen) {
-					return "arguments", node, node.Lparen, node.Rparen
-				}
-			case *ast.CompositeLit: // T{a, b, c}
-				if isCursorInside(node.Lbrace, node.Rbrace) {
-					return "elements", node, node.Lbrace, node.Rbrace
+		cur, _ := curFile.FindByPos(start, end)
+		for cur := range cur.Enclosing() {
+			// TODO: do cur = enclosingUnparen(cur) first, once CL 701035 lands.
+			switch cur.ParentEdgeKind() {
+			// params or results of func signature
+			// Note:
+			// - each ast.Field (e.g. "x, y, z int") is considered a single item.
+			// - splitting Params and Results lists is not usually good style.
+			case edge.FuncType_Params:
+				p := cur.Node().(*ast.FieldList)
+				return "parameters", p, p.Opening, p.Closing
+			case edge.FuncType_Results:
+				r := cur.Node().(*ast.FieldList)
+				if !r.Opening.IsValid() {
+					continue
 				}
+				return "results", r, r.Opening, r.Closing
+			case edge.CallExpr_Args: // f(a, b, c)
+				node := cur.Parent().Node().(*ast.CallExpr)
+				return "arguments", node, node.Lparen, node.Rparen
+			case edge.CompositeLit_Elts: // T{a, b, c}
+				node := cur.Parent().Node().(*ast.CompositeLit)
+				return "elements", node, node.Lbrace, node.Rbrace
 			}
 		}
-
 		return "", nil, 0, 0
 	}
 
@@ -205,6 +215,18 @@ func findSplitJoinTarget(fset *token.FileSet, file *ast.File, src []byte, start,
 		for _, arg := range node.Args {
 			items = append(items, arg)
 		}
+
+		// Preserve "..." by wrapping the last
+		// argument in an Ellipsis node
+		// with the same Pos/End as the argument.
+		// See corresponding logic in processLines.
+		if node.Ellipsis.IsValid() {
+			last := &items[len(items)-1]
+			*last = &ast.Ellipsis{
+				Ellipsis: (*last).Pos(),      // determines Ellipsis.Pos()
+				Elt:      (*last).(ast.Expr), // determines Ellipsis.End()
+			}
+		}
 	case *ast.CompositeLit:
 		for _, arg := range node.Elts {
 			items = append(items, arg)
@@ -212,6 +234,7 @@ func findSplitJoinTarget(fset *token.FileSet, file *ast.File, src []byte, start,
 	}
 
 	// preserve comments separately as it's not part of the targetNode AST.
+	file := curFile.Node().(*ast.File)
 	for _, cg := range file.Comments {
 		if open <= cg.Pos() && cg.Pos() < close {
 			comments = append(comments, cg)
