@@ -14,7 +14,7 @@ package golang
 // - rewrite using html/template.
 //   Or factor with golang.org/x/pkgsite/internal/godoc/dochtml.
 // - emit breadcrumbs for parent + sibling packages.
-// - list promoted methods---we have type information!
+// - list promoted methods---we have type information! (golang/go#67158)
 // - gather Example tests, following go/doc and pkgsite.
 // - add option for doc.AllDecls: show non-exported symbols too.
 // - style the <li> bullets in the index as invisible.
@@ -40,17 +40,15 @@ import (
 	"go/types"
 	"html"
 	"path/filepath"
+	"slices"
 	"strings"
 
-	"golang.org/x/tools/go/ast/astutil"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/protocol"
-	goplsastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
-	"golang.org/x/tools/gopls/internal/util/safetoken"
-	"golang.org/x/tools/gopls/internal/util/slices"
-	"golang.org/x/tools/gopls/internal/util/typesutil"
+	"golang.org/x/tools/gopls/internal/util/cursorutil"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/stdlib"
 	"golang.org/x/tools/internal/typesinternal"
 )
@@ -80,7 +78,7 @@ func DocFragment(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (p
 		// External test packages don't have /pkg doc pages,
 		// so instead show the doc for the package under test.
 		// (This named-based heuristic is imperfect.)
-		if forTest := strings.TrimSuffix(pkg.Path(), "_test"); forTest != pkg.Path() {
+		if forTest, ok := strings.CutSuffix(pkg.Path(), "_test"); ok {
 			return PackagePath(forTest), "", makeTitle("package", nil, filepath.Base(forTest))
 		}
 
@@ -120,7 +118,7 @@ func DocFragment(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (p
 	if !sym.Exported() {
 		// Unexported method of exported type?
 		if fn, ok := sym.(*types.Func); ok {
-			if recv := fn.Type().(*types.Signature).Recv(); recv != nil {
+			if recv := fn.Signature().Recv(); recv != nil {
 				_, named := typesinternal.ReceiverNamed(recv)
 				if named != nil && named.Obj().Exported() {
 					sym = named.Obj()
@@ -140,14 +138,14 @@ func DocFragment(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) (p
 	}
 
 	// package-level symbol?
-	if isPackageLevel(sym) {
+	if typesinternal.IsPackageLevel(sym) {
 		return pkgpath, sym.Name(), makeTitle(objectKind(sym), sym.Pkg(), sym.Name())
 	}
 
 	// Inv: sym is field or method, or local.
 	switch sym := sym.(type) {
 	case *types.Func: // => method
-		sig := sym.Type().(*types.Signature)
+		sig := sym.Signature()
 		isPtr, named := typesinternal.ReceiverNamed(sig.Recv())
 		if named != nil {
 			if !named.Obj().Exported() {
@@ -194,20 +192,18 @@ type thing struct {
 }
 
 func thingAtPoint(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) thing {
-	path, _ := astutil.PathEnclosingInterval(pgf.File, start, end)
+	cur, _, _, _ := astutil.Select(pgf.Cursor(), start, end)
 
 	// In an import spec?
-	if len(path) >= 3 { // [...ImportSpec GenDecl File]
-		if spec, ok := path[len(path)-3].(*ast.ImportSpec); ok {
-			if pkgname, ok := typesutil.ImportedPkgName(pkg.TypesInfo(), spec); ok {
-				return thing{pkg: pkgname.Imported()}
-			}
+	if spec, _ := cursorutil.FirstEnclosing[*ast.ImportSpec](cur); spec != nil {
+		if pkgname := pkg.TypesInfo().PkgNameOf(spec); pkgname != nil {
+			return thing{pkg: pkgname.Imported()}
 		}
 	}
 
 	// Definition or reference to symbol?
 	var obj types.Object
-	if id, ok := path[0].(*ast.Ident); ok {
+	if id, ok := cur.Node().(*ast.Ident); ok {
 		obj = pkg.TypesInfo().ObjectOf(id)
 
 		// Treat use to PkgName like ImportSpec.
@@ -215,7 +211,7 @@ func thingAtPoint(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) t
 			return thing{pkg: pkgname.Imported()}
 		}
 
-	} else if sel, ok := path[0].(*ast.SelectorExpr); ok {
+	} else if sel, ok := cur.Node().(*ast.SelectorExpr); ok {
 		// e.g. selection is "fmt.Println" or just a portion ("mt.Prin")
 		obj = pkg.TypesInfo().Uses[sel.Sel]
 	}
@@ -223,35 +219,21 @@ func thingAtPoint(pkg *cache.Package, pgf *parsego.File, start, end token.Pos) t
 		return thing{symbol: obj}
 	}
 
-	// Find enclosing declaration.
-	if n := len(path); n > 1 {
-		switch decl := path[n-2].(type) {
-		case *ast.FuncDecl:
-			// method?
-			if fn := pkg.TypesInfo().Defs[decl.Name]; fn != nil {
-				return thing{enclosing: fn}
-			}
-
-		case *ast.GenDecl:
-			// path=[... Spec? GenDecl File]
-			for _, spec := range decl.Specs {
-				if n > 2 && spec == path[n-3] {
-					var name *ast.Ident
-					switch spec := spec.(type) {
-					case *ast.ValueSpec:
-						// var, const: use first name
-						name = spec.Names[0]
-					case *ast.TypeSpec:
-						name = spec.Name
-					}
-					if name != nil {
-						return thing{enclosing: pkg.TypesInfo().Defs[name]}
-					}
-					break
-				}
-			}
+	// Find outermost enclosing declaration.
+	var id *ast.Ident
+	for cur := range cur.Enclosing((*ast.FuncDecl)(nil), (*ast.ValueSpec)(nil), (*ast.TypeSpec)(nil)) {
+		switch n := cur.Node().(type) {
+		case *ast.FuncDecl: // function or method
+			id = n.Name
+		case *ast.ValueSpec:
+			id = n.Names[0] // var, const: use first name
+		case *ast.TypeSpec:
+			id = n.Name
 		}
 	}
+	if obj := pkg.TypesInfo().Defs[id]; obj != nil {
+		return thing{enclosing: obj}
+	}
 
 	return thing{} // nothing to see here
 }
@@ -270,6 +252,13 @@ type Web interface {
 // The posURL function returns a URL that when visited, has the side
 // effect of causing gopls to direct the client editor to navigate to
 // the specified file/line/column position, in UTF-8 coordinates.
+//
+// TODO(adonovan): this function could use some unit tests; we
+// shouldn't have to use integration tests to cover microdetails of
+// HTML rendering. (It is tempting to abstract this function so that
+// it depends only on FileSet/File/Types/TypeInfo/etc, but we should
+// bend the tests to the production interfaces, not the other way
+// around.)
 func PackageDocHTML(viewID string, pkg *cache.Package, web Web) ([]byte, error) {
 	// We can't use doc.NewFromFiles (even with doc.PreserveAST
 	// mode) as it calls ast.NewPackage which assumes that each
@@ -282,7 +271,7 @@ func PackageDocHTML(viewID string, pkg *cache.Package, web Web) ([]byte, error)
 	// TODO(adonovan): simulate that too.
 	fileMap := make(map[string]*ast.File)
 	for _, f := range pkg.Syntax() {
-		fileMap[pkg.FileSet().File(f.Pos()).Name()] = f
+		fileMap[pkg.FileSet().File(f.FileStart).Name()] = f
 	}
 	astpkg := &ast.Package{
 		Name:  pkg.Types().Name(),
@@ -322,72 +311,48 @@ func PackageDocHTML(viewID string, pkg *cache.Package, web Web) ([]byte, error)
 			filterValues(&t.Vars)
 			filterFuncs(&t.Funcs)
 			filterFuncs(&t.Methods)
-			return unexported(t.Name)
+			if unexported(t.Name) {
+				// If an unexported type has an exported constructor function,
+				// treat the constructor as an ordinary standalone function.
+				// We will sort Funcs again below.
+				docpkg.Funcs = append(docpkg.Funcs, t.Funcs...)
+				return true // delete this type
+			}
+			return false // keep this type
+		})
+		slices.SortFunc(docpkg.Funcs, func(x, y *doc.Func) int {
+			return strings.Compare(x.Name, y.Name)
 		})
 	}
 
-	var docHTML func(comment string) []byte
+	// docHTML renders the doc comment as Markdown.
+	// The fileNode is used to deduce the enclosing file
+	// for the correct import mapping.
+	//
+	// It is not concurrency-safe.
+	var docHTML func(fileNode ast.Node, comment string) []byte
 	{
 		// Adapt doc comment parser and printer
 		// to our representation of Go packages
 		// so that doc links (e.g. "[fmt.Println]")
 		// become valid links.
-
-		printer := docpkg.Printer()
-		printer.DocLinkURL = func(link *comment.DocLink) string {
-			path := pkg.Metadata().PkgPath
-			if link.ImportPath != "" {
-				path = PackagePath(link.ImportPath)
-			}
-			fragment := link.Name
-			if link.Recv != "" {
-				fragment = link.Recv + "." + link.Name
-			}
-			return web.PkgURL(viewID, path, fragment)
-		}
-		parser := docpkg.Parser()
-		parser.LookupPackage = func(name string) (importPath string, ok bool) {
-			// Ambiguous: different files in the same
-			// package may have different import mappings,
-			// but the hook doesn't provide the file context.
-			// TODO(adonovan): conspire with docHTML to
-			// pass the doc comment's enclosing file through
-			// a shared variable, so that we can compute
-			// the correct per-file mapping.
-			//
-			// TODO(adonovan): check for PkgName.Name
-			// matches, but also check for
-			// PkgName.Imported.Namer matches, since some
-			// packages are typically imported under a
-			// non-default name (e.g. pathpkg "path") but
-			// may be referred to in doc links using their
-			// canonical name.
-			for _, f := range pkg.Syntax() {
-				for _, imp := range f.Imports {
-					pkgName, ok := typesutil.ImportedPkgName(pkg.TypesInfo(), imp)
-					if ok && pkgName.Name() == name {
-						return pkgName.Imported().Path(), true
-					}
+		printer := &comment.Printer{
+			DocLinkURL: func(link *comment.DocLink) string {
+				path := pkg.Metadata().PkgPath
+				if link.ImportPath != "" {
+					path = PackagePath(link.ImportPath)
 				}
-			}
-			return "", false
+				fragment := link.Name
+				if link.Recv != "" {
+					fragment = link.Recv + "." + link.Name
+				}
+				return web.PkgURL(viewID, path, fragment)
+			},
 		}
-		parser.LookupSym = func(recv, name string) (ok bool) {
-			// package-level decl?
-			if recv == "" {
-				return pkg.Types().Scope().Lookup(name) != nil
-			}
-
-			// method?
-			tname, ok := pkg.Types().Scope().Lookup(recv).(*types.TypeName)
-			if !ok {
-				return false
-			}
-			m, _, _ := types.LookupFieldOrMethod(tname.Type(), true, pkg.Types(), name)
-			return is[*types.Func](m)
-		}
-		docHTML = func(comment string) []byte {
-			return printer.HTML(parser.Parse(comment))
+		parse := newDocCommentParser(pkg)
+		docHTML = func(fileNode ast.Node, comment string) []byte {
+			doc := parse(fileNode, comment)
+			return printer.HTML(doc)
 		}
 	}
 
@@ -469,7 +434,7 @@ window.addEventListener('load', function() {
 		label := obj.Name() // for a type
 		if fn, ok := obj.(*types.Func); ok {
 			var buf strings.Builder
-			sig := fn.Type().(*types.Signature)
+			sig := fn.Signature()
 			if sig.Recv() != nil {
 				fmt.Fprintf(&buf, "(%s) ", sig.Recv().Name())
 				fragment = recvType + "." + fn.Name()
@@ -551,7 +516,7 @@ window.addEventListener('load', function() {
 
 				// method of package-level named type?
 				if fn, ok := obj.(*types.Func); ok {
-					sig := fn.Type().(*types.Signature)
+					sig := fn.Signature()
 					if sig.Recv() != nil {
 						_, named := typesinternal.ReceiverNamed(sig.Recv())
 						if named != nil {
@@ -585,7 +550,7 @@ window.addEventListener('load', function() {
 		// appear as separate decls. We should too.
 		var buf bytes.Buffer
 		for _, file := range pkg.CompiledGoFiles() {
-			if goplsastutil.NodeContains(file.File, n.Pos()) {
+			if astutil.NodeContainsPos(file.File, n.Pos()) {
 				pos := n.Pos()
 
 				// emit emits source in the interval [pos:to] and updates pos.
@@ -595,15 +560,12 @@ window.addEventListener('load', function() {
 					if !to.IsValid() {
 						bug.Reportf("invalid Pos")
 					}
-					start, err := safetoken.Offset(file.Tok, pos)
+					text, err := file.PosText(pos, to)
 					if err != nil {
-						bug.Reportf("invalid start Pos: %v", err)
+						bug.Reportf("invalid pos range: %v", err)
+						return
 					}
-					end, err := safetoken.Offset(file.Tok, to)
-					if err != nil {
-						bug.Reportf("invalid end Pos: %v", err)
-					}
-					buf.WriteString(escape(string(file.Src[start:end])))
+					buf.WriteString(escape(string(text)))
 					pos = to
 				}
 				ast.Inspect(n, func(n ast.Node) bool {
@@ -648,7 +610,7 @@ window.addEventListener('load', function() {
 	fnString := func(fn *types.Func) string {
 		pkgRelative := typesinternal.NameRelativeTo(pkg.Types())
 
-		sig := fn.Type().(*types.Signature)
+		sig := fn.Signature()
 
 		// Emit "func (recv T) F".
 		var buf bytes.Buffer
@@ -693,8 +655,8 @@ window.addEventListener('load', function() {
 				cloneTparams(sig.RecvTypeParams()),
 				cloneTparams(sig.TypeParams()),
 				types.NewTuple(append(
-					typesSeqToSlice[*types.Var](sig.Params())[:3],
-					types.NewVar(0, nil, "", types.Typ[types.Invalid]))...),
+					slices.Collect(sig.Params().Variables())[:3],
+					types.NewParam(0, nil, "", types.Typ[types.Invalid]))...),
 				sig.Results(),
 				false) // any final ...T parameter is truncated
 		}
@@ -715,7 +677,12 @@ window.addEventListener('load', function() {
 		"https://pkg.go.dev/"+string(pkg.Types().Path()))
 
 	// package doc
-	fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(docpkg.Doc))
+	for _, f := range pkg.Syntax() {
+		if f.Doc != nil {
+			fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(f.Doc, docpkg.Doc))
+			break
+		}
+	}
 
 	// symbol index
 	fmt.Fprintf(&buf, "<h2 id='hdr-Index'>Index</h2>\n")
@@ -773,7 +740,7 @@ window.addEventListener('load', function() {
 			fmt.Fprintf(&buf, "<pre class='code'>%s</pre>\n", nodeHTML(&decl2))
 
 			// comment (if any)
-			fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(v.Doc))
+			fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(v.Decl, v.Doc))
 		}
 	}
 	fmt.Fprintf(&buf, "<h2 id='hdr-Constants'>Constants</h2>\n")
@@ -814,7 +781,7 @@ window.addEventListener('load', function() {
 				nodeHTML(docfn.Decl.Type))
 
 			// comment (if any)
-			fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(docfn.Doc))
+			fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(docfn.Decl, docfn.Doc))
 		}
 	}
 	funcs(docpkg.Funcs)
@@ -835,7 +802,7 @@ window.addEventListener('load', function() {
 		fmt.Fprintf(&buf, "<pre class='code'>%s</pre>\n", nodeHTML(&decl2))
 
 		// comment (if any)
-		fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(doctype.Doc))
+		fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n", docHTML(doctype.Decl, doctype.Doc))
 
 		// subelements
 		values(doctype.Consts) // constants of type T
@@ -856,7 +823,7 @@ window.addEventListener('load', function() {
 
 			// comment (if any)
 			fmt.Fprintf(&buf, "<div class='comment'>%s</div>\n",
-				docHTML(docmethod.Doc))
+				docHTML(docmethod.Decl, docmethod.Doc))
 		}
 	}
 
@@ -873,19 +840,3 @@ window.addEventListener('load', function() {
 
 	return buf.Bytes(), nil
 }
-
-// typesSeq abstracts various go/types sequence types:
-// MethodSet, Tuple, TypeParamList, TypeList.
-// TODO(adonovan): replace with go1.23 iterators.
-type typesSeq[T any] interface {
-	Len() int
-	At(int) T
-}
-
-func typesSeqToSlice[T any](seq typesSeq[T]) []T {
-	slice := make([]T, seq.Len())
-	for i := range slice {
-		slice[i] = seq.At(i)
-	}
-	return slice
-}
