@@ -24,18 +24,21 @@ import (
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/debug"
 	debuglog "golang.org/x/tools/gopls/internal/debug/log"
-	"golang.org/x/tools/gopls/internal/file"
+	"golang.org/x/tools/gopls/internal/filecache"
 	"golang.org/x/tools/gopls/internal/protocol"
+	"golang.org/x/tools/gopls/internal/protocol/semtok"
 	"golang.org/x/tools/gopls/internal/settings"
+	"golang.org/x/tools/gopls/internal/telemetry"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/goversion"
-	"golang.org/x/tools/gopls/internal/util/maps"
+	"golang.org/x/tools/gopls/internal/util/moremaps"
+	"golang.org/x/tools/gopls/internal/util/moreslices"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/jsonrpc2"
 )
 
 func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitialize) (*protocol.InitializeResult, error) {
-	ctx, done := event.Start(ctx, "lsp.Server.initialize")
+	ctx, done := event.Start(ctx, "server.Initialize")
 	defer done()
 
 	var clientName string
@@ -52,29 +55,23 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 	s.state = serverInitializing
 	s.stateMu.Unlock()
 
-	// For uniqueness, use the gopls PID rather than params.ProcessID (the client
-	// pid). Some clients might start multiple gopls servers, though they
-	// probably shouldn't.
-	pid := os.Getpid()
-	s.tempDir = filepath.Join(os.TempDir(), fmt.Sprintf("gopls-%d.%s", pid, s.session.ID()))
-	err := os.Mkdir(s.tempDir, 0700)
-	if err != nil {
-		// MkdirTemp could fail due to permissions issues. This is a problem with
-		// the user's environment, but should not block gopls otherwise behaving.
-		// All usage of s.tempDir should be predicated on having a non-empty
-		// s.tempDir.
-		event.Error(ctx, "creating temp dir", err)
-		s.tempDir = ""
-	}
 	s.progress.SetSupportsWorkDoneProgress(params.Capabilities.Window.WorkDoneProgress)
 
 	options := s.Options().Clone()
 	// TODO(rfindley): eliminate this defer.
 	defer func() { s.SetOptions(options) }()
 
-	s.handleOptionErrors(ctx, options.Set(params.InitializationOptions))
+	// Process initialization options.
+	{
+		res, errs := options.Set(params.InitializationOptions)
+		s.handleOptionResult(ctx, res, errs)
+	}
 	options.ForClientCapabilities(params.ClientInfo, params.Capabilities)
 
+	if options.MaxFileCacheBytes > 0 {
+		filecache.SetBudget(options.MaxFileCacheBytes)
+	}
+
 	if options.ShowBugReports {
 		// Report the next bug that occurs on the server.
 		bug.Handle(func(b bug.Bug) {
@@ -97,7 +94,7 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 	}
 	s.pendingFolders = append(s.pendingFolders, folders...)
 
-	var codeActionProvider interface{} = true
+	var codeActionProvider any = true
 	if ca := params.Capabilities.TextDocument.CodeAction; len(ca.CodeActionLiteralSupport.CodeActionKind.ValueSet) > 0 {
 		// If the client has specified CodeActionLiteralSupport,
 		// send the code actions we support.
@@ -108,13 +105,36 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 			ResolveProvider: true,
 		}
 	}
-	var renameOpts interface{} = true
+
+	var diagnosticProvider *protocol.Or_ServerCapabilities_diagnosticProvider
+	if options.PullDiagnostics {
+		diagnosticProvider = &protocol.Or_ServerCapabilities_diagnosticProvider{
+			Value: protocol.DiagnosticOptions{
+				InterFileDependencies: true,
+				WorkspaceDiagnostics:  false, // we don't support workspace/diagnostic
+			},
+		}
+	}
+
+	var renameOpts any = true
 	if r := params.Capabilities.TextDocument.Rename; r != nil && r.PrepareSupport {
 		renameOpts = protocol.RenameOptions{
 			PrepareProvider: r.PrepareSupport,
 		}
 	}
 
+	var semanticTokenProvider any
+	if options.SemanticTokens {
+		semanticTokenProvider = protocol.SemanticTokensOptions{
+			Range: &protocol.Or_SemanticTokensOptions_range{Value: true},
+			Full:  &protocol.Or_SemanticTokensOptions_full{Value: true},
+			Legend: protocol.SemanticTokensLegend{
+				TokenTypes:     moreslices.ConvertStrings[string](semtok.Types),
+				TokenModifiers: moreslices.ConvertStrings[string](semtok.Modifiers),
+			},
+		}
+	}
+
 	versionInfo := debug.VersionInfo()
 
 	goplsVersion, err := json.Marshal(versionInfo)
@@ -144,19 +164,16 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 			DocumentHighlightProvider: &protocol.Or_ServerCapabilities_documentHighlightProvider{Value: true},
 			DocumentLinkProvider:      &protocol.DocumentLinkOptions{},
 			InlayHintProvider:         protocol.InlayHintOptions{},
+			DiagnosticProvider:        diagnosticProvider,
 			ReferencesProvider:        &protocol.Or_ServerCapabilities_referencesProvider{Value: true},
 			RenameProvider:            renameOpts,
 			SelectionRangeProvider:    &protocol.Or_ServerCapabilities_selectionRangeProvider{Value: true},
-			SemanticTokensProvider: protocol.SemanticTokensOptions{
-				Range: &protocol.Or_SemanticTokensOptions_range{Value: true},
-				Full:  &protocol.Or_SemanticTokensOptions_full{Value: true},
-				Legend: protocol.SemanticTokensLegend{
-					TokenTypes:     protocol.NonNilSlice(options.SemanticTypes),
-					TokenModifiers: protocol.NonNilSlice(options.SemanticMods),
-				},
-			},
+			SemanticTokensProvider:    semanticTokenProvider,
 			SignatureHelpProvider: &protocol.SignatureHelpOptions{
 				TriggerCharacters: []string{"(", ","},
+				// Used to update or dismiss signature help when it's already active,
+				// typically after a call expression is closed.
+				RetriggerCharacters: []string{")"},
 			},
 			TextDocumentSync: &protocol.TextDocumentSyncOptions{
 				Change:    protocol.Incremental,
@@ -165,11 +182,37 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 					IncludeText: false,
 				},
 			},
+			TypeHierarchyProvider: &protocol.Or_ServerCapabilities_typeHierarchyProvider{Value: true},
 			Workspace: &protocol.WorkspaceOptions{
 				WorkspaceFolders: &protocol.WorkspaceFolders5Gn{
 					Supported:           true,
 					ChangeNotifications: "workspace/didChangeWorkspaceFolders",
 				},
+				FileOperations: &protocol.FileOperationOptions{
+					DidCreate: &protocol.FileOperationRegistrationOptions{
+						Filters: []protocol.FileOperationFilter{{
+							Scheme: "file",
+							// gopls is only interested with files in .go extension.
+							Pattern: protocol.FileOperationPattern{Glob: "**/*.go"},
+						}},
+					},
+				},
+			},
+			Experimental: map[string]any{
+				// interactiveResolveProvider lists the LSP objects that support
+				// an interactive resolution stage. For instance, the presence of
+				// "command" indicates that the server handles "command/resolve"
+				// requests.
+				//
+				// Due to the existence of "codeAction/resolve" and language
+				// clients that resolve code action eagerly, "codeAction" can
+				// never be interactively resolved.
+				//
+				// TODO(hxjiang): experiment with interactively resolving
+				// "RenameParams". See golang/go#69107.
+				//
+				// TODO(pjw): add "command" to the supported list.
+				"interactiveResolveProvider": []string{},
 			},
 		},
 		ServerInfo: &protocol.ServerInfo{
@@ -180,7 +223,7 @@ func (s *server) Initialize(ctx context.Context, params *protocol.ParamInitializ
 }
 
 func (s *server) Initialized(ctx context.Context, params *protocol.InitializedParams) error {
-	ctx, done := event.Start(ctx, "lsp.Server.initialized")
+	ctx, done := event.Start(ctx, "server.Initialized")
 	defer done()
 
 	s.stateMu.Lock()
@@ -192,7 +235,7 @@ func (s *server) Initialized(ctx context.Context, params *protocol.InitializedPa
 	s.stateMu.Unlock()
 
 	for _, not := range s.notifications {
-		s.client.ShowMessage(ctx, not)
+		s.client.ShowMessage(ctx, not) // ignore error
 	}
 	s.notifications = nil
 
@@ -284,10 +327,15 @@ func (s *server) addFolders(ctx context.Context, folders []protocol.WorkspaceFol
 	// but the list can grow over time.
 	var filtered []protocol.WorkspaceFolder
 	for _, f := range folders {
-		if _, err := protocol.ParseDocumentURI(f.URI); err != nil {
+		uri, err := protocol.ParseDocumentURI(f.URI)
+		if err != nil {
 			debuglog.Warning.Logf(ctx, "skip adding virtual folder %q - invalid folder URI: %v", f.Name, err)
 			continue
 		}
+		if s.session.HasView(uri) {
+			debuglog.Warning.Logf(ctx, "skip adding the already added folder %q - its view has been created before", f.Name)
+			continue
+		}
 		filtered = append(filtered, f)
 	}
 	folders = filtered
@@ -333,13 +381,11 @@ func (s *server) addFolders(ctx context.Context, folders []protocol.WorkspaceFol
 		}()
 
 		// Diagnose the newly created view asynchronously.
-		ndiagnose.Add(1)
-		go func() {
+		ndiagnose.Go(func() {
 			s.diagnoseSnapshot(snapshot.BackgroundContext(), snapshot, nil, 0)
 			<-initialized
 			release()
-			ndiagnose.Done()
-		}()
+		})
 	}
 
 	// Wait for snapshots to be initialized so that all files are known.
@@ -353,11 +399,12 @@ func (s *server) addFolders(ctx context.Context, folders []protocol.WorkspaceFol
 
 	// Report any errors using the protocol.
 	if len(viewErrors) > 0 {
-		errMsg := fmt.Sprintf("Error loading workspace folders (expected %v, got %v)\n", len(folders), len(s.session.Views())-originalViews)
+		var errMsg strings.Builder
+		fmt.Fprintf(&errMsg, "Error loading workspace folders (expected %v, got %v)\n", len(folders), len(s.session.Views())-originalViews)
 		for uri, err := range viewErrors {
-			errMsg += fmt.Sprintf("failed to load view for %s: %v\n", uri, err)
+			fmt.Fprintf(&errMsg, "failed to load view for %s: %v\n", uri, err)
 		}
-		showMessage(ctx, s.client, protocol.Error, errMsg)
+		showMessage(ctx, s.client, protocol.Error, errMsg.String())
 	}
 }
 
@@ -372,7 +419,7 @@ func (s *server) updateWatchedDirectories(ctx context.Context) error {
 	defer s.watchedGlobPatternsMu.Unlock()
 
 	// Nothing to do if the set of workspace directories is unchanged.
-	if maps.SameKeys(s.watchedGlobPatterns, patterns) {
+	if moremaps.SameKeys(s.watchedGlobPatterns, patterns) {
 		return nil
 	}
 
@@ -472,13 +519,24 @@ func (s *server) newFolder(ctx context.Context, folder protocol.DocumentURI, nam
 	// Increment folder counters.
 	switch {
 	case env.GOTOOLCHAIN == "auto" || strings.Contains(env.GOTOOLCHAIN, "+auto"):
-		counter.New("gopls/gotoolchain:auto").Inc()
+		counter.Inc("gopls/gotoolchain:auto")
 	case env.GOTOOLCHAIN == "path" || strings.Contains(env.GOTOOLCHAIN, "+path"):
-		counter.New("gopls/gotoolchain:path").Inc()
+		counter.Inc("gopls/gotoolchain:path")
 	case env.GOTOOLCHAIN == "local": // local+auto and local+path handled above
-		counter.New("gopls/gotoolchain:local").Inc()
+		counter.Inc("gopls/gotoolchain:local")
 	default:
-		counter.New("gopls/gotoolchain:other").Inc()
+		counter.Inc("gopls/gotoolchain:other")
+	}
+
+	// Record whether a driver is in use so that it appears in the
+	// user's telemetry upload. Although we can't correlate the
+	// driver information with the crash or bug.Report at the
+	// granularity of the process instance, users that use a
+	// driver tend to do so most of the time, so we'll get a
+	// strong clue. See #60890 for an example of an issue where
+	// this information would have been helpful.
+	if env.EffectiveGOPACKAGESDRIVER != "" {
+		counter.Inc("gopls/gopackagesdriver")
 	}
 
 	return &cache.Folder{
@@ -516,7 +574,8 @@ func (s *server) fetchFolderOptions(ctx context.Context, folder protocol.Documen
 
 	opts = opts.Clone()
 	for _, config := range configs {
-		s.handleOptionErrors(ctx, opts.Set(config))
+		res, errs := opts.Set(config)
+		s.handleOptionResult(ctx, res, errs)
 	}
 	return opts, nil
 }
@@ -530,7 +589,12 @@ func (s *server) eventuallyShowMessage(ctx context.Context, msg *protocol.ShowMe
 	s.notifications = append(s.notifications, msg)
 }
 
-func (s *server) handleOptionErrors(ctx context.Context, optionErrors []error) {
+func (s *server) handleOptionResult(ctx context.Context, applied []telemetry.CounterPath, optionErrors []error) {
+	for _, path := range applied {
+		path = append(settings.CounterPath{"gopls", "setting"}, path...)
+		counter.Inc(path.FullName())
+	}
+
 	var warnings, errs []string
 	for _, err := range optionErrors {
 		if err == nil {
@@ -572,25 +636,10 @@ func (s *server) handleOptionErrors(ctx context.Context, optionErrors []error) {
 	}
 }
 
-// fileOf returns the file for a given URI and its snapshot.
-// On success, the returned function must be called to release the snapshot.
-func (s *server) fileOf(ctx context.Context, uri protocol.DocumentURI) (file.Handle, *cache.Snapshot, func(), error) {
-	snapshot, release, err := s.session.SnapshotOf(ctx, uri)
-	if err != nil {
-		return nil, nil, nil, err
-	}
-	fh, err := snapshot.ReadFile(ctx, uri)
-	if err != nil {
-		release()
-		return nil, nil, nil, err
-	}
-	return fh, snapshot, release, nil
-}
-
-// shutdown implements the 'shutdown' LSP handler. It releases resources
+// Shutdown implements the 'shutdown' LSP handler. It releases resources
 // associated with the server and waits for all ongoing work to complete.
 func (s *server) Shutdown(ctx context.Context) error {
-	ctx, done := event.Start(ctx, "lsp.Server.shutdown")
+	ctx, done := event.Start(ctx, "server.Shutdown")
 	defer done()
 
 	s.stateMu.Lock()
@@ -601,29 +650,24 @@ func (s *server) Shutdown(ctx context.Context) error {
 	if s.state != serverShutDown {
 		// Wait for the webserver (if any) to finish.
 		if s.web != nil {
-			s.web.server.Shutdown(ctx)
+			s.web.server.Shutdown(ctx) // ignore error
 		}
 
 		// drop all the active views
 		s.session.Shutdown(ctx)
 		s.state = serverShutDown
-		if s.tempDir != "" {
-			if err := os.RemoveAll(s.tempDir); err != nil {
-				event.Error(ctx, "removing temp dir", err)
-			}
-		}
 	}
 	return nil
 }
 
 func (s *server) Exit(ctx context.Context) error {
-	ctx, done := event.Start(ctx, "lsp.Server.exit")
+	ctx, done := event.Start(ctx, "server.Exit")
 	defer done()
 
 	s.stateMu.Lock()
 	defer s.stateMu.Unlock()
 
-	s.client.Close()
+	s.client.Close() // ignore error
 
 	if s.state != serverShutDown {
 		// TODO: We should be able to do better than this.
@@ -653,6 +697,9 @@ func recordClientInfo(clientName string) {
 	case "govim":
 		// https://github.com/govim/govim/pull/1189
 		key = "gopls/client:govim"
+	case "helix":
+		// https://github.com/helix-editor/helix/blob/d0218f7e78bc0c3af4b0995ab8bda66b9c542cf3/helix-lsp/src/client.rs#L714
+		key = "gopls/client:helix"
 	case "Neovim":
 		// https://github.com/neovim/neovim/blob/42333ea98dfcd2994ee128a3467dfe68205154cd/runtime/lua/vim/lsp.lua#L1361
 		key = "gopls/client:neovim"
@@ -662,6 +709,13 @@ func recordClientInfo(clientName string) {
 	case "Sublime Text LSP":
 		// https://github.com/sublimelsp/LSP/blob/e608f878e7e9dd34aabe4ff0462540fadcd88fcc/plugin/core/sessions.py#L493
 		key = "gopls/client:sublimetext"
+	case "Windsurf":
+		key = "gopls/client:windsurf"
+	case "Cursor":
+		key = "gopls/client:cursor"
+	case "Zed", "Zed Dev", "Zed Nightly", "Zed Preview":
+		// https: //github.com/zed-industries/zed/blob/0ac17526687bf11007f0fbb5c3b2ff463ce47293/crates/release_channel/src/lib.rs#L147
+		key = "gopls/client:zed"
 	default:
 		// Accumulate at least a local counter for an unknown
 		// client name, but also fall through to count it as
