@@ -12,6 +12,7 @@ import (
 	"golang.org/x/tools/gopls/internal/label"
 	"golang.org/x/tools/gopls/internal/mod"
 	"golang.org/x/tools/gopls/internal/protocol"
+	"golang.org/x/tools/gopls/internal/settings"
 	"golang.org/x/tools/gopls/internal/telemetry"
 	"golang.org/x/tools/gopls/internal/template"
 	"golang.org/x/tools/gopls/internal/work"
@@ -24,10 +25,10 @@ func (s *server) Hover(ctx context.Context, params *protocol.HoverParams) (_ *pr
 		recordLatency(ctx, rerr)
 	}()
 
-	ctx, done := event.Start(ctx, "lsp.Server.hover", label.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "server.Hover", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
-	fh, snapshot, release, err := s.fileOf(ctx, params.TextDocument.URI)
+	fh, snapshot, release, err := s.session.FileOf(ctx, params.TextDocument.URI)
 	if err != nil {
 		return nil, err
 	}
@@ -35,13 +36,24 @@ func (s *server) Hover(ctx context.Context, params *protocol.HoverParams) (_ *pr
 
 	switch snapshot.FileKind(fh) {
 	case file.Mod:
-		return mod.Hover(ctx, snapshot, fh, params.Position)
+		return mod.Hover(ctx, snapshot, fh, params.Range)
 	case file.Go:
-		return golang.Hover(ctx, snapshot, fh, params.Position)
+		var pkgURL func(path golang.PackagePath, fragment string) protocol.URI
+		if snapshot.Options().LinksInHover == settings.LinksInHover_Gopls {
+			web, err := s.getWeb()
+			if err != nil {
+				event.Error(ctx, "failed to start web server", err)
+			} else {
+				pkgURL = func(path golang.PackagePath, fragment string) protocol.URI {
+					return web.PkgURL(snapshot.View().ID(), path, fragment)
+				}
+			}
+		}
+		return golang.Hover(ctx, snapshot, fh, params.Range, pkgURL)
 	case file.Tmpl:
-		return template.Hover(ctx, snapshot, fh, params.Position)
+		return template.Hover(ctx, snapshot, fh, params.Range)
 	case file.Work:
-		return work.Hover(ctx, snapshot, fh, params.Position)
+		return work.Hover(ctx, snapshot, fh, params.Range)
 	}
 	return nil, nil // empty result
 }
