@@ -10,6 +10,7 @@ import (
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/golang"
 	"golang.org/x/tools/gopls/internal/label"
+	"golang.org/x/tools/gopls/internal/mod"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/telemetry"
 	"golang.org/x/tools/gopls/internal/template"
@@ -22,10 +23,10 @@ func (s *server) References(ctx context.Context, params *protocol.ReferenceParam
 		recordLatency(ctx, rerr)
 	}()
 
-	ctx, done := event.Start(ctx, "lsp.Server.references", label.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "server.References", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
-	fh, snapshot, release, err := s.fileOf(ctx, params.TextDocument.URI)
+	fh, snapshot, release, err := s.session.FileOf(ctx, params.TextDocument.URI)
 	if err != nil {
 		return nil, err
 	}
@@ -34,7 +35,9 @@ func (s *server) References(ctx context.Context, params *protocol.ReferenceParam
 	case file.Tmpl:
 		return template.References(ctx, snapshot, fh, params)
 	case file.Go:
-		return golang.References(ctx, snapshot, fh, params.Position, params.Context.IncludeDeclaration)
+		return golang.References(ctx, snapshot, fh, params.Range, params.Context.IncludeDeclaration)
+	case file.Mod:
+		return mod.References(ctx, snapshot, fh, params)
 	}
 	return nil, nil // empty result
 }
