@@ -61,9 +61,9 @@ const (
 	// ResetGoModDiagnostics command.
 	FromResetGoModDiagnostics
 
-	// FromToggleGCDetails refers to state changes resulting from toggling
-	// gc_details on or off for a package.
-	FromToggleGCDetails
+	// FromToggleCompilerOptDetails refers to state changes resulting from toggling
+	// a package's compiler optimization details flag.
+	FromToggleCompilerOptDetails
 )
 
 func (m ModificationSource) String() string {
@@ -92,7 +92,7 @@ func (m ModificationSource) String() string {
 }
 
 func (s *server) DidOpen(ctx context.Context, params *protocol.DidOpenTextDocumentParams) error {
-	ctx, done := event.Start(ctx, "lsp.Server.didOpen", label.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "server.DidOpen", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
 	uri := params.TextDocument.URI
@@ -105,7 +105,7 @@ func (s *server) DidOpen(ctx context.Context, params *protocol.DidOpenTextDocume
 	// file is opened, and we can't do that inside didModifyFiles because we
 	// don't want to request configuration while holding a lock.
 	if len(s.session.Views()) == 0 {
-		dir := filepath.Dir(uri.Path())
+		dir := uri.DirPath()
 		s.addFolders(ctx, []protocol.WorkspaceFolder{{
 			URI:  string(protocol.URIFromPath(dir)),
 			Name: filepath.Base(dir),
@@ -121,7 +121,7 @@ func (s *server) DidOpen(ctx context.Context, params *protocol.DidOpenTextDocume
 }
 
 func (s *server) DidChange(ctx context.Context, params *protocol.DidChangeTextDocumentParams) error {
-	ctx, done := event.Start(ctx, "lsp.Server.didChange", label.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "server.DidChange", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
 	uri := params.TextDocument.URI
@@ -156,25 +156,26 @@ func (s *server) warnAboutModifyingGeneratedFiles(ctx context.Context, uri proto
 		return nil
 	}
 
-	// Ideally, we should be able to specify that a generated file should
-	// be opened as read-only. Tell the user that they should not be
-	// editing a generated file.
+	// Warn the user that they are editing a generated file, but
+	// don't try to stop them: there are often good reasons to do
+	// so, such as adding temporary logging, or evaluating changes
+	// to the generated code without the trouble of modifying the
+	// generator logic (see #73959).
 	snapshot, release, err := s.session.SnapshotOf(ctx, uri)
 	if err != nil {
 		return err
 	}
 	isGenerated := golang.IsGenerated(ctx, snapshot, uri)
 	release()
-
 	if isGenerated {
-		msg := fmt.Sprintf("Do not edit this file! %s is a generated file.", uri.Path())
+		msg := fmt.Sprintf("Warning: editing %s, a generated file.", uri.Base())
 		showMessage(ctx, s.client, protocol.Warning, msg)
 	}
 	return nil
 }
 
 func (s *server) DidChangeWatchedFiles(ctx context.Context, params *protocol.DidChangeWatchedFilesParams) error {
-	ctx, done := event.Start(ctx, "lsp.Server.didChangeWatchedFiles")
+	ctx, done := event.Start(ctx, "server.DidChangeWatchedFiles")
 	defer done()
 
 	var modifications []file.Modification
@@ -190,7 +191,7 @@ func (s *server) DidChangeWatchedFiles(ctx context.Context, params *protocol.Did
 }
 
 func (s *server) DidSave(ctx context.Context, params *protocol.DidSaveTextDocumentParams) error {
-	ctx, done := event.Start(ctx, "lsp.Server.didSave", label.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "server.DidSave", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
 	c := file.Modification{
@@ -204,7 +205,7 @@ func (s *server) DidSave(ctx context.Context, params *protocol.DidSaveTextDocume
 }
 
 func (s *server) DidClose(ctx context.Context, params *protocol.DidCloseTextDocumentParams) error {
-	ctx, done := event.Start(ctx, "lsp.Server.didClose", label.URI.Of(params.TextDocument.URI))
+	ctx, done := event.Start(ctx, "server.DidClose", label.URI.Of(params.TextDocument.URI))
 	defer done()
 
 	return s.didModifyFiles(ctx, []file.Modification{
@@ -253,6 +254,17 @@ func (s *server) didModifyFiles(ctx context.Context, modifications []file.Modifi
 	// to their files.
 	modifications = s.session.ExpandModificationsToDirectories(ctx, modifications)
 
+	// TODO: also handle go.work changes as well.
+	uris := make(map[protocol.DocumentURI]struct{})
+	for _, m := range modifications {
+		if strings.HasSuffix(m.URI.Path(), "go.mod") && (m.Action == file.Create || m.Action == file.Save) {
+			uris[m.URI] = struct{}{}
+		}
+	}
+	for uri := range uris {
+		go s.checkGoModDeps(ctx, uri)
+	}
+
 	viewsToDiagnose, err := s.session.DidModifyFiles(ctx, modifications)
 	if err != nil {
 		return err
@@ -265,11 +277,9 @@ func (s *server) didModifyFiles(ctx context.Context, modifications []file.Modifi
 
 	modCtx, modID := s.needsDiagnosis(ctx, viewsToDiagnose)
 
-	wg.Add(1)
-	go func() {
+	wg.Go(func() {
 		s.diagnoseChangedViews(modCtx, modID, viewsToDiagnose, cause)
-		wg.Done()
-	}()
+	})
 
 	// After any file modifications, we need to update our watched files,
 	// in case something changed. Compute the new set of directories to watch,
@@ -313,7 +323,7 @@ func (s *server) changedText(ctx context.Context, uri protocol.DocumentURI, chan
 
 	// Check if the client sent the full content of the file.
 	// We accept a full content change even if the server expected incremental changes.
-	if len(changes) == 1 && changes[0].Range == nil && changes[0].RangeLength == 0 {
+	if len(changes) == 1 && changes[0].Range == nil && changes[0].RangeLength == nil {
 		changeFull.Inc()
 		return []byte(changes[0].Text), nil
 	}
@@ -388,7 +398,7 @@ func (s *server) checkEfficacy(uri protocol.DocumentURI, version int32, change p
 		}
 		if edit.Range.Start == change.Range.Start {
 			// the change and the proposed completion start at the same
-			if change.RangeLength == 0 && len(change.Text) == 1 {
+			if (change.RangeLength == nil || *change.RangeLength == 0) && len(change.Text) == 1 {
 				// a single character added it does not count as a completion
 				continue
 			}
