@@ -3,7 +3,6 @@
 // license that can be found in the LICENSE file.
 
 //go:build go1.21 && !openbsd && !js && !wasip1 && !solaris && !android && !386
-// +build go1.21,!openbsd,!js,!wasip1,!solaris,!android,!386
 
 package telemetry_test
 
@@ -26,13 +25,13 @@ import (
 )
 
 func TestMain(m *testing.M) {
-	tmp, err := os.MkdirTemp("", "gopls-telemetry-test")
+	tmp, err := os.MkdirTemp("", "gopls-telemetry-test-counters")
 	if err != nil {
 		panic(err)
 	}
 	countertest.Open(tmp)
 	code := Main(m)
-	os.RemoveAll(tmp)
+	os.RemoveAll(tmp) // ignore error (cleanup fails on Windows; golang/go#68243)
 	os.Exit(code)
 }
 
@@ -54,6 +53,7 @@ func TestTelemetry(t *testing.T) {
 		counter.New("gopls/client:" + editor),
 		counter.New("gopls/goversion:1." + goversion),
 		counter.New("fwd/vscode/linter:a"),
+		counter.New("gopls/gotoolchain:local"),
 	}
 	initialCounts := make([]uint64, len(sessionCounters))
 	for i, c := range sessionCounters {
@@ -70,6 +70,9 @@ func TestTelemetry(t *testing.T) {
 		Modes(Default), // must be in-process to receive the bug report below
 		Settings{"showBugReports": true},
 		ClientName("Visual Studio Code"),
+		EnvVars{
+			"GOTOOLCHAIN": "local", // so that the local counter is incremented
+		},
 	).Run(t, "", func(_ *testing.T, env *Env) {
 		goversion = strconv.Itoa(env.GoVersion())
 		addForwardedCounters(env, []string{"vscode/linter:a"}, []int64{1})
@@ -93,6 +96,7 @@ func TestTelemetry(t *testing.T) {
 	// gopls/editor:client
 	// gopls/goversion:1.x
 	// fwd/vscode/linter:a
+	// gopls/gotoolchain:local
 	for i, c := range sessionCounters {
 		want := initialCounts[i] + 1
 		got, err := countertest.ReadCounter(c)
@@ -114,20 +118,64 @@ func TestTelemetry(t *testing.T) {
 	}
 }
 
+func TestSettingTelemetry(t *testing.T) {
+	// counters that should be incremented by each session
+	sessionCounters := []*counter.Counter{
+		counter.New("gopls/setting/diagnosticsDelay"),
+		counter.New("gopls/setting/staticcheck:true"),
+		counter.New("gopls/setting/noSemanticString:true"),
+		counter.New("gopls/setting/analyses/deprecated:false"),
+	}
+
+	initialCounts := make([]uint64, len(sessionCounters))
+	for i, c := range sessionCounters {
+		count, err := countertest.ReadCounter(c)
+		if err != nil {
+			continue // counter db not open, or counter not found
+		}
+		initialCounts[i] = count
+	}
+
+	// Run gopls.
+	WithOptions(
+		Modes(Default),
+		Settings{
+			"staticcheck": true,
+			"analyses": map[string]bool{
+				"deprecated": false,
+			},
+			"diagnosticsDelay": "0s",
+			"noSemanticString": true,
+		},
+	).Run(t, "", func(_ *testing.T, env *Env) {
+	})
+
+	for i, c := range sessionCounters {
+		count, err := countertest.ReadCounter(c)
+		if err != nil {
+			t.Errorf("ReadCounter(%q) failed: %v", c.Name(), err)
+			continue
+		}
+		if count <= initialCounts[i] {
+			t.Errorf("ReadCounter(%q) = %d, want > %d", c.Name(), count, initialCounts[i])
+		}
+	}
+}
+
 func addForwardedCounters(env *Env, names []string, values []int64) {
 	args, err := command.MarshalArgs(command.AddTelemetryCountersArgs{
 		Names: names, Values: values,
 	})
 	if err != nil {
-		env.T.Fatal(err)
+		env.TB.Fatal(err)
 	}
 	var res error
 	env.ExecuteCommand(&protocol.ExecuteCommandParams{
-		Command:   command.AddTelemetryCounters.ID(),
+		Command:   command.AddTelemetryCounters.String(),
 		Arguments: args,
 	}, &res)
 	if res != nil {
-		env.T.Errorf("%v failed - %v", command.AddTelemetryCounters.ID(), res)
+		env.TB.Errorf("%v failed - %v", command.AddTelemetryCounters, res)
 	}
 }
 
@@ -215,7 +263,7 @@ func _() {
 	).Run(t, files, func(_ *testing.T, env *Env) {
 		env.OpenFile("a.go")
 		before := totalLatencySamples(t, "completion", false)
-		loc := env.RegexpSearch("a.go", "x")
+		loc := env.RegexpSearch("a.go", "()x")
 		for i := 0; i < 10; i++ {
 			env.Completion(loc)
 		}
