@@ -4,11 +4,15 @@
 
 package completion
 
+// This file is misnamed; it has no version constraints.
+// TODO(adonovan):  fold into completion_test.go
+
 import (
 	"testing"
 
 	"golang.org/x/tools/gopls/internal/protocol"
 	. "golang.org/x/tools/gopls/internal/test/integration"
+	"golang.org/x/tools/internal/testenv"
 )
 
 // test generic receivers
@@ -52,14 +56,18 @@ func (s SyncMap[XX,string]) g(v UU) {}
 		}
 	})
 }
+
 func TestFuzzFunc(t *testing.T) {
-	t.Skip("broken on release-branch.0.25 for unknown reasons")
+	// The behavior under test is derived from the std module,
+	// not the x/tools/internal/stdlib linked into gopls.
+	testenv.NeedsGoCommand1Point(t, 25) // go1.25 added TBF.Attr
+
 	// use the example from the package documentation
 	modfile := `
 -- go.mod --
 module mod.com
 
-go 1.18
+go 1.25
 `
 	part0 := `package foo
 import "testing"
@@ -92,15 +100,16 @@ func FuzzHex(f *testing.F) {
 ` + part0 + part1 + part2
 
 	tests := []struct {
-		file   string
-		pat    string
-		offset uint32 // UTF16 length from the beginning of pat to what the user just typed
-		want   []string
+		file string
+		pat  string
+		want []string
 	}{
-		{"a_test.go", "f.Ad", 3, []string{"Add"}},
-		{"c_test.go", " f.F", 4, []string{"Failed"}},
-		{"c_test.go", "f.N", 3, []string{"Name"}},
-		{"b_test.go", "f.F", 3, []string{"Fuzz(func(t *testing.T, a []byte)", "Fail", "FailNow",
+		// To avoid breaking these assertions as the "testing" package evolves,
+		// use an optional (?) suffix for newer symbols.
+		{"a_test.go", "f.A()d", []string{"Add", "ArtifactDir?", "Attr"}}, // Attr is 1.25, Artifact is 1.26
+		{"c_test.go", " f.F()", []string{"Failed"}},
+		{"c_test.go", "f.N()", []string{"Name"}},
+		{"b_test.go", "f.F()", []string{"Fuzz(func(t *testing.T, a []byte)", "Fail", "FailNow",
 			"Failed", "Fatal", "Fatalf"}},
 	}
 	Run(t, data, func(t *testing.T, env *Env) {
@@ -108,11 +117,10 @@ func FuzzHex(f *testing.F) {
 			env.OpenFile(test.file)
 			env.Await(env.DoneWithOpen())
 			loc := env.RegexpSearch(test.file, test.pat)
-			loc.Range.Start.Character += test.offset // character user just typed? will type?
 			completions := env.Completion(loc)
 			result := compareCompletionLabels(test.want, completions.Items)
 			if result != "" {
-				t.Errorf("pat %q %q", test.pat, result)
+				t.Errorf("pat=%q <<%s>>", test.pat, result)
 				for i, it := range completions.Items {
 					t.Errorf("%d got %q %q", i, it.Label, it.Detail)
 				}
